package Analysis;
import java.util.ArrayList;
import Common.Constants;
import Common.Enums;
import Common.Task;


public class MulticontrollerAnalysis 
{
	private ArrayList<Task> taskset;
	private Enums.ActionResult schedulable; 
	private long noOfCores; 
	private long givenMemoryBudgetContrOne, givenMemoryBudgetContrTwo; 

	public MulticontrollerAnalysis(ArrayList<Task> ts, long memorybudgetContrOne, long memorybudgetContrTwo, long cores)
	{		
		noOfCores = cores; 
		//givenMemoryBudgetContrOne = memorybudgetContrOne; // Q1 
		//givenMemoryBudgetContrTwo = memorybudgetContrTwo; // Q2

		givenMemoryBudgetContrOne = Math.max(0, Constants.period - ((long)Math.ceil((double)(Constants.period - memorybudgetContrOne)/(double)(noOfCores - 1)) * (noOfCores - 1))); 
		givenMemoryBudgetContrTwo = Math.max(0,Constants.period - ((long)Math.ceil((double)(Constants.period - memorybudgetContrTwo)/(double)(noOfCores - 1)) * (noOfCores - 1))); 

		taskset = new ArrayList<Task>(ts.size()); 
		taskset = ts; 
		schedulable = Enums.ActionResult.Failure; 

		if(givenMemoryBudgetContrOne == 0 || givenMemoryBudgetContrTwo == 0) // this loop check if we have budget zero with some memory accesses than we should return failure
		{
			for(Task task: taskset)
			{
				if( (task.GetM_WCET(0)> 0 && givenMemoryBudgetContrOne ==0) ||  (task.GetM_WCET(1)> 0 && givenMemoryBudgetContrTwo ==0))
					return; 
			}
		}

		schedulable = AssignPriorities(); 
	}


	private Enums.ActionResult AssignPriorities() 
	{

		boolean flag = false; 

		for(Task task:taskset)
			task.SetPriority(-1);

		for(int k=taskset.size()-1; k>=0; k--)
		{
			flag = false; 
			for(Task task:taskset)
			{
				if(task.GetPriority() == -1)
				{
					task.SetPriority(k);
					if(PerformYaoAnalysis(task) == Enums.ActionResult.Failure)
					{
						task.SetPriority(-1);
					}
					else
					{
						flag = true; 
						break; 
					}
				}
			}

			if(flag == false)
				return Enums.ActionResult.Failure; 
		}
		return Enums.ActionResult.Success; 
	}


	private Enums.ActionResult PerformYaoAnalysis(Task task) 
	{

		long responseTimeWithoutStall = 0;  
		long responseTimeWithStall = 0;  

		responseTimeWithoutStall = ComputeResponseTime(task, 0, 0); 
		if(responseTimeWithoutStall > (long)task.GetDeadline())
			return Enums.ActionResult.Failure;



		responseTimeWithStall = ComputeResponseTime(task, responseTimeWithoutStall, 1); 
		if(responseTimeWithStall > (long)task.GetDeadline() )
			return Enums.ActionResult.Failure;

		// if all the tests are passed then pass success
		return Enums.ActionResult.Success; 

	}

	private long[] ComputeDemand(long time, Task task) 
	{
		long[] totaldemand = {0,0,0,0}; 

		long jobs = 0; 

		for(Task t: taskset) 
		{
			jobs = (long) Math.ceil((double)time/(double)t.GetMinInterArrival()); 

			if(t.GetPriority()<=task.GetPriority())
			{
				if(!task.equals(t))
					totaldemand[0] += (jobs * (long)t.GetWCET()); 

				totaldemand[1] += (jobs * (long)t.GetE_WCET()); 
				totaldemand[2] += (jobs * (long)t.GetM_WCET(0)); 
				totaldemand[3] += (jobs * (long)t.GetM_WCET(1));
			}

		}

		return totaldemand;
	}

	private long ComputeResponseTime(Task task, long startResponseTime, int stall) 
	{
		long prevResponseTime =-1, currResponseTime = startResponseTime; 
		long [] response = {0,0,0};  

		//long old_analysis, new_analysis; 

		while(currResponseTime != prevResponseTime && currResponseTime <=task.GetDeadline())
		{
			prevResponseTime = currResponseTime;
			response = ComputeDemand(currResponseTime,task); 
			currResponseTime = (long)task.GetWCET() + response[0];

			if(stall == 1) //Response time with stall
			{
				currResponseTime += Stall(response[1],response[2],response[3],response[1]+response[2]+response[3]);

				/*old_analysis = Stall_old(response[1],response[2],response[3],response[1]+response[2]+response[3]);
				new_analysis = Stall(response[1],response[2],response[3],response[1]+response[2]+response[3]);
				new_analysis = Stall(response[1],response[2],response[3],response[1]+response[2]+response[3]);

				if(old_analysis != new_analysis)
				{
					//System.out.println("The results of two analysis is not same.\t" + "Diff:\t" + (old_analysis - new_analysis) +"\t Old Analysis:\t"+old_analysis + "\t New Analysis:\t" +new_analysis);
					old_analysis = Stall_old(response[1],response[2],response[3],response[1]+response[2]+response[3]);
					new_analysis = Stall(response[1],response[2],response[3],response[1]+response[2]+response[3]);
				}

				stall_old = currResponseTime+Stall_old(response[1],response[2],response[3],response[1]+response[2]+response[3]);
				 */

			}
		}

		return currResponseTime;
	}



	/******************New Analysis Functions*******************/

	private long Stall(long eWCET, long m1WCET, long m2WCET, long WCET)
	{
		long stall = 0;
		long Q1 =0 , Q2 = 0; 
		double b1 =0.0, b2 =0.0; 

		Q1 = givenMemoryBudgetContrOne;
		Q2 = givenMemoryBudgetContrTwo; 

		b1 = (double)Q1/(double)Constants.period;
		b2 = (double)Q2/(double)Constants.period;


		double RBS1 = (long)((double)(Constants.period-Q1)/(double)(noOfCores - 1)); 
		double RBS2 = (long)((double)(Constants.period-Q2)/(double)(noOfCores - 1));



		if(m1WCET == 0 && m2WCET == 0)
			return 0; 
		else if(m1WCET == 0)
			stall = YaoStall(eWCET, m2WCET, Q2);
		else if(m2WCET == 0)
			stall = YaoStall(eWCET, m1WCET, Q1); 
		else if(RBS1 == 0 && RBS2 == 0)
			return 0; 
		else if (RBS1 == 0)
			stall = YaoStall(eWCET + m1WCET, m2WCET, Q2); 
		else if (RBS2 == 0)
			stall = YaoStall(eWCET + m2WCET, m1WCET, Q1);
		// Regulation dominant case
		else if(b1 <= (double)1/(double)noOfCores && b2 <= (double)1/(double)noOfCores)
		{ 
			stall = YaoStall(0, (long) (Math.floor((double)m1WCET/(double)Q1)*Q1), Q1)+
					YaoStall(0, (long) (Math.floor((double)m2WCET/(double)Q2)*Q2), Q2) +
					( ((m1WCET % Q1) + (m2WCET % Q2))*(noOfCores-1)); 
		}
		// Contention dominant case
		else if(b1 > (double)1/(double)noOfCores && b2 > (double)1/(double)noOfCores)
		{
			long cm1 =0, cm2=0; 

			if(m2WCET * (long)RBS1 > m1WCET * (long)RBS2)
				cm2 = (long)RBS2; 
			else 
				cm1 = (long)RBS1; 

			if(cm1 == (long)RBS1)
			{
				if((long)RBS1 * m2WCET < (long)((Constants.period/noOfCores) - RBS1) * m1WCET)
				{
					stall = CaseTwoStall(Q1, cm1, m1WCET, RBS1, m2WCET, RBS2, eWCET);
				}
				else 
				{
					stall = (m1WCET + m2WCET ) * (noOfCores - 1); 
				}
			}
			else if(cm2 == (RBS2))
			{
				if(RBS2 * m1WCET < (long)((Constants.period/noOfCores)-RBS2) * m2WCET)
				{		
					stall = CaseTwoStall(Q2, cm2, m2WCET, RBS2, m1WCET, RBS1, eWCET);
				}
				else 
				{
					stall = (m1WCET + m2WCET ) * (noOfCores - 1); 
				}
			}
			else if (cm1 != RBS1 && cm2 !=	RBS2)

			{
				stall = (m1WCET + m2WCET ) * (noOfCores - 1); // FIRST CASE
			}
			else
			{
				System.out.println("Something wrong in new analysis 2nd case:\tcm1:" + cm1 +"\tRBS1:"+RBS1+"\tcm2:"+cm2+"\tRBS2:"+RBS2+"\tP/m:"+(Constants.period/noOfCores)+"\tTolerance:"+ Constants.roundingTolerance);
			}


			/******************************old***************************/
			/*double rm = (double)m2WCET/(double)m1WCET; 
			//long cm1  =  (long)Math.min(RBS1, Math.min(RBS2/rm, (double)Constants.period/(double)(noOfCores * (1+rm)))); 
			double cm1  =  Math.min(RBS1, Math.min((RBS2*(double)m1WCET)/(double)m2WCET, (double)(Constants.period * m1WCET)/(double)(noOfCores * (m1WCET+m2WCET))));
			double cm2  =  (rm * cm1);  


			if(WithinRoundingTolerance((long)(cm1+cm2), (double)Constants.period/(double)noOfCores))// || ( ((long)(cm1+cm2) >= (long)(((double)Constants.period/(double)noOfCores)-staticRound)) && ((long)(cm1+cm2) <= (long)(((double)Constants.period/(double)noOfCores)+staticRound))))       //if( (cm1 + cm2 == (long)((double)Constants.period/(double)noOfCores)) || ( (cm1 + cm2 >= (long)(((double)Constants.period/(double)noOfCores)- Constants.toleranceInRounding)) && (cm1 + cm2 <= (long)(((double)Constants.period/(double)noOfCores)+ Constants.toleranceInRounding))) ) // missing case
			{
				stall = (m1WCET + m2WCET ) * (noOfCores - 1); 
			}
			else if(WithinRoundingTolerance((long)cm1,RBS1))// || ((long)cm1 >= (long)(RBS1-staticRound) && (long)cm1<= (long)(RBS1+staticRound)))       //else if( (cm1 == (long)RBS1) || ( (cm1 >= (long)(RBS1-RoundingTolerance(RBS1))) && (cm1 <=  (long)(RBS1+Constants.toleranceInRounding)) )) // cm1* = RBS1, the second value cm1 == (long)RBS1 is added just to avoid those very few case that does not get the exact value of RBS1 due rounding of error
			{
				stall = CaseTwoStall(Q1, (long)cm1, m1WCET, RBS1, m2WCET, RBS2, eWCET);
			}
			else if (WithinRoundingTolerance((long)cm2,RBS2))// || ((long)cm2 >= (long)(RBS2-staticRound) && (long)cm2<= (long)(RBS2+staticRound)))     	//else if ( (cm2 == (long)RBS2) || ((cm2 >= (long)(RBS2-Constants.toleranceInRounding)) && (cm2 <=  (long)(RBS2+Constants.toleranceInRounding)) )) //cm2* = RBS2, the second value cm2 == (long)RBS2 is added just to avoid those very few case that does not get the exact value of RBS1 due rounding of error
			{
				stall = CaseTwoStall(Q2, (long)cm2, m2WCET, RBS2, m1WCET, RBS1, eWCET);
			}
			else
				System.out.println("Something wrong in new analysis 2nd case:\tcm1:" + cm1 +"\tRBS1:"+RBS1+"\tcm2:"+cm2+"\tRBS2:"+RBS2+"\tP/m:"+(Constants.period/noOfCores)+"\tTolerance:"+ Constants.roundingTolerance);
			 */
		}
		// neither contention nor regulation strictly dominant 
		else 
		{

			if(b1 <= (double)1/(double)noOfCores && b2 > (double)1/(double)noOfCores)
			{
				stall = CaseThreeStall(Q1, m1WCET, Q2, m2WCET, RBS2, eWCET);	
			}
			else if(b2 <= (double)1/(double)noOfCores && b1 > (double)1/(double)noOfCores)
			{
				stall = CaseThreeStall(Q2, m2WCET, Q1, m1WCET, RBS1, eWCET);
			}
			else 
			{
				System.out.println("In neither contention nor regulation strictly domiant case but none of the two possible (b1 <= 1/noOfCores && b2 > 1/noOfCores) or (b2 <= 1/noOfCores && b1 > 1/noOfCores) exists.");
			}
		}
		stall += (Constants.period - Math.min(Q1, Q2)); // add the first regulationperiod  
		return stall ; 

	}

	private long CaseTwoStall(long Q1, long cm1_asterisk, long m1WCET, double RBS1, long m2WCET, double RBS2, long eWCET)
	{
		long stall = 0; 
		double rm = (double)m2WCET/(double)m1WCET; 
		long cm2_asterisk = (long)Math.max(rm * (double)cm1_asterisk, Math.min(RBS2, ((double)Constants.period/(double)noOfCores)- RBS1)); 
		long K1_asterisk = Math.min((long)Math.floor((double)m2WCET/(double)cm2_asterisk),(long)Math.floor((double)(eWCET+m1WCET+m2WCET)/(double)(Constants.period - ((cm1_asterisk+cm2_asterisk)*(noOfCores-1))) ) ); // equation 35  

		long stall_1 = K1_asterisk * (cm1_asterisk + cm2_asterisk) * (noOfCores - 1); // equation 11


		long eWCET_cap = Math.max(0, eWCET - (K1_asterisk * (Constants.period - ((cm1_asterisk + cm2_asterisk)*noOfCores)))); // equation 14

		long mWCET_cap = m1WCET + m2WCET + eWCET - (K1_asterisk * (Constants.period - ((cm1_asterisk + cm2_asterisk)*(noOfCores - 1)))); 


		long m2WCET_cap = 0; 
		if(eWCET >= K1_asterisk *(Constants.period - ((cm1_asterisk+cm2_asterisk)*noOfCores)))
		{
			m2WCET_cap = m2WCET - (K1_asterisk * cm2_asterisk) ;
		}
		else if( (K1_asterisk == (long)Math.floor((double)m2WCET/(double)cm2_asterisk)) && (eWCET < K1_asterisk *(Constants.period - ((cm1_asterisk+cm2_asterisk)*noOfCores))))
		{
			m2WCET_cap = Math.min(m2WCET - (K1_asterisk * cm2_asterisk), mWCET_cap); 			
		}
		else 
		{
			m2WCET_cap = Math.min(m2WCET - (K1_asterisk * cm2_asterisk), mWCET_cap - Math.min(m1WCET - (K1_asterisk * cm1_asterisk), Math.min((long)RBS1, mWCET_cap))); 
		}


		long m1WCET_cap = 0;  // equation 38

		if(eWCET >= K1_asterisk *(Constants.period - ((cm1_asterisk+cm2_asterisk)*noOfCores)))
		{
			m1WCET_cap = m1WCET  - (K1_asterisk * cm1_asterisk); 
		}
		else 
		{
			m1WCET_cap = mWCET_cap - m2WCET_cap; 
		}

		long stall_23 = YaoStall(m2WCET_cap + (Math.min(m2WCET_cap, (long)RBS2) * (noOfCores-1)) + eWCET_cap, m1WCET_cap, Q1);

		stall = stall_1 + ((noOfCores - 1)* Math.min(m2WCET_cap, (long)RBS2)) 
				+ stall_23 ; 


		if(m1WCET_cap < 0 || m2WCET_cap <0 || mWCET_cap < 0)
			System.out.println("some problem in case 2 analysis.");

		return stall; 
	}


	private long CaseThreeStall(long Q1,long m1WCET, long Q2, long m2WCET, double RBS2, long eWCET) 
	{

		long stall = 0, deltarho = 0; 

		long deltaCe = noOfCores * Q1; 

		long deltaCcm2_max =  (long)((RBS2 * (long)Math.floor((double)deltaCe/(double)Q2)) + Math.min(RBS2, deltaCe % Q2)); ; // equation 43; 
		long deltaCcm2_min =  (long)((RBS2 * (long)Math.floor((double)Math.max(0, deltaCe - (Constants.period - (RBS2*noOfCores)) )/(double)Q2)) + Math.min(RBS2, Math.max(0, deltaCe - (Constants.period - (RBS2*noOfCores))) % Q2)); ; // equation 43 but replace deltaCe with max(deltaCe - ( P - m * RBS2), 0)
		long deltaCcm2_asterisk = (long)Math.floor((double)(Constants.period - (noOfCores * Q1))/(double)(noOfCores - 1)); 		

		if(eWCET >= (long)( (Math.floor((double)m2WCET/RBS2)*(Constants.period - (noOfCores * RBS2))) - ((m1WCET%Q1) * noOfCores)) )
		{
			deltarho = 0; 
		}
		else if(deltaCcm2_max <= deltaCcm2_asterisk)
		{
			deltarho = 0; 
		}
		else if(deltaCcm2_min > deltaCcm2_asterisk)
		{
			long stall_1 = YaoStall(eWCET + ((m1WCET%Q1)*noOfCores), m2WCET, Q2); 
			long R = m2WCET + eWCET + ((m1WCET%Q1) * noOfCores) + stall_1; 
			long Ccm2_cap = m2WCET - (long)(Math.floor((double)R/(double)Constants.period)*RBS2) - (long)Math.min( Math.floor( (double)(R % Constants.period) / (double)noOfCores), RBS2); 
			long deltarho_temp = 0 , m1WCET_cap = 0, Ccm2_cap_temp = 0; 

			while(Ccm2_cap > deltaCcm2_asterisk && deltarho < (long)Math.floor((double)m1WCET/(double)Q1))
			{

				deltarho_temp = deltarho + 1; 

				m1WCET_cap = (m1WCET % Q1 ) + (deltarho_temp * Q1) ; 
				stall_1 = YaoStall(eWCET + (m1WCET_cap * noOfCores), m2WCET, Q2); 
				R = m2WCET + eWCET + (m1WCET_cap * noOfCores) + stall_1;
				Ccm2_cap_temp = Math.max(0, m2WCET - (long)(Math.floor((double)R/(double)Constants.period)*RBS2) - (long)Math.min(Math.floor((double)(R % Constants.period)/(double)noOfCores), RBS2) ); 

				if( (m1WCET_cap - Math.min(Q1-1, Math.max(0, (long)Math.floor((double)(R%Constants.period)/(double)noOfCores)-RBS2)) )  <  (long)( (Q1-1) * Math.floor((double)R/(double)Constants.period)) )
				{
					deltarho = deltarho_temp; 
					Ccm2_cap = Ccm2_cap_temp; 
				}
				else 
				{
					break; 
				}

			}

		}
		else 
		{
			long deltarho_max = 0, stall_max = 0, stall_1 = 0, R = 0;//, Cm1_cap = 0; 

			long m1WCET_cap = 0; 
			for(deltarho = 0; deltarho <= (long)Math.floor((double)m1WCET/(double)Q1); deltarho++)
			{
				m1WCET_cap = (m1WCET % Q1 ) + (deltarho * Q1) ; 

				stall_1 = YaoStall(eWCET + (m1WCET_cap * noOfCores), m2WCET, Q2); 
				R = m2WCET + eWCET + (m1WCET_cap * noOfCores) + stall_1;

				if ((stall_1 +  ((long)(Math.floor((double)m1WCET/(double)Q1) - deltarho) * (Constants.period - Q1))) > stall_max && 
						(m1WCET_cap - Math.min(Q1-1, Math.max(0, (long)Math.floor((double)(R%Constants.period)/(double)noOfCores)-RBS2))) < (long)((Q1 - 1) *  Math.floor((double)R/(double)Constants.period)) )
				{
					stall_max = stall_1 + ((long)(Math.floor((double)m1WCET/(double)Q1) - deltarho) * (Constants.period - Q1)) ;  
					deltarho_max = deltarho; 
				}
			}

			deltarho = deltarho_max; 	
		}


		stall = YaoStall(0, m1WCET - (m1WCET % Q1) - (deltarho * Q1), Q1) + 
				YaoStall(eWCET + (((m1WCET%Q1)+ (deltarho * Q1))*noOfCores), m2WCET, Q2) + 
				(((m1WCET %Q1) + (deltarho * Q1)) * (noOfCores - 1)); 
		return stall;
	}

	private boolean WithinRoundingTolerance(long leftValue, double rightValue)
	{
		long low = (long)Math.floor(rightValue - (rightValue*Constants.roundingTolerance)); 
		long high= (long)Math.ceil(rightValue + (rightValue*Constants.roundingTolerance)); 

		if(leftValue >= low && leftValue <= high)
			return true;
		else 
			return false; 
	}

	public Enums.ActionResult GetSchedulabilityStatus()
	{
		return schedulable; 
	}

	public int GetMaximumDeadline() 
	{

		int maximumdeadline = 0; 		
		for(Task task: taskset)
		{
			if(maximumdeadline<task.GetDeadline())
				maximumdeadline = task.GetDeadline(); 
		}

		return maximumdeadline;
	}



	/************************************ old analysis *******************************************/
	private long Stall_old(long eWCET, long m1WCET, long m2WCET, long WCET)
	{
		long stall = 0;
		long Q1 =0 , Q2 = 0; 
		double b1 =0.0, b2 =0.0; 

		Q1 = givenMemoryBudgetContrOne;
		Q2 = givenMemoryBudgetContrTwo; 

		b1 = (double)Q1/(double)Constants.period;
		b2 = (double)Q2/(double)Constants.period;


		double RBS1 = (long)((double)(Constants.period-Q1)/(double)(noOfCores - 1)); 
		double RBS2 = (long)((double)(Constants.period-Q2)/(double)(noOfCores - 1));


		if(m1WCET == 0 && m2WCET == 0)
			return 0; 
		else if(m1WCET == 0)
			stall = YaoStall(eWCET, m2WCET, Q2);
		else if(m2WCET == 0)
			stall = YaoStall(eWCET, m1WCET, Q1); 
		else if(RBS1 == 0 && RBS2 == 0)
			return 0; 
		else if (RBS1 == 0)
			stall = YaoStall(eWCET + m1WCET, m2WCET, Q2); 
		else if (RBS2 == 0)
			stall = YaoStall(eWCET + m2WCET, m1WCET, Q1);
		// Regulation dominant case
		else if(b1 <= (double)1/(double)noOfCores && b2 <= (double)1/(double)noOfCores)
		{
			stall =  (long)(Math.floor((double)m1WCET/(double)Q1)*(Constants.period - Q1)) + ((m1WCET % Q1)*(noOfCores -1)) +
					(long)(Math.floor((double)m2WCET/(double)Q2)*(Constants.period - Q2)) + ((m2WCET % Q2)*(noOfCores -1)) ; 

			//System.out.println("Old Analysis:\t" +  ((long)(Math.floor((double)m1WCET/(double)Q1)*(Constants.period - Q1)) + ((m1WCET % Q1)*(noOfCores -1))) + "\t" + ((long)(Math.floor((double)m2WCET/(double)Q2)*(Constants.period - Q2)) + ((m2WCET % Q2)*(noOfCores -1))) + "\t" + stall); 

		}
		// Contention dominant case
		else if(b1 > (double)1/(double)noOfCores && b2 > (double)1/(double)noOfCores)
		{
			double rm = (double)m2WCET/(double)m1WCET; 
			double cm1  =  Math.min(RBS1, Math.min(RBS2/rm, (double)Constants.period/(double)(noOfCores * (1+rm)))); 
			double cm2  =  (rm * cm1);  

			if((long)cm1 != (long)RBS1 && (long)cm2 != (long)RBS2) // missing case
			{
				stall = (m1WCET + m2WCET ) * (noOfCores - 1); 
			}
			else if(WithinRoundingTolerance((long)cm1,RBS1)) // || (cm1 >= (RBS1-staticRound) && cm1<= (RBS1+staticRound))) //	else if(WithinRoundingTolerance(cm1,RBS1)) //else if((cm1 == (long)RBS1) || ((cm1 >= (long)(RBS1-Constants.toleranceInRounding)) && (cm1 <=  (long)(RBS1+Constants.toleranceInRounding)) ))  			//else if(cm1 == (long)RBS1) // cm1* = RBS1 
			{
				stall = CaseTwoStall_old(Q1, (long)cm1, m1WCET, RBS1, m2WCET, RBS2, eWCET, WCET); 
			}
			else if (WithinRoundingTolerance((long)cm2,RBS2)) //|| (cm2 >= (RBS2-staticRound) && cm2<= (RBS2+staticRound))) //else if (WithinRoundingTolerance(cm2,RBS2))   //else if ((cm2 == (long)RBS2) || ( (cm2 >= (long)(RBS2-Constants.toleranceInRounding)) && (cm2 <=  (long)(RBS2+Constants.toleranceInRounding)) )) 
			{
				stall = CaseTwoStall_old(Q2, (long)cm2, m2WCET, RBS2, m1WCET, RBS1, eWCET, WCET); 
			}
			else 
				System.out.println("Something wrong in the 2nd case of old analysis");

		}
		// neither contention nor regulation strictly dominant 
		else 
		{

			if(b1 <= (double)1/(double)noOfCores && b2 > (double)1/(double)noOfCores)
			{
				stall = CaseThreeStall_old(b1, Q1, m1WCET, RBS1, b2, Q2, m2WCET, RBS2, eWCET);
			}
			else if(b2 <= (double)1/(double)noOfCores && b1 > (double)1/(double)noOfCores)
			{
				stall = CaseThreeStall_old(b2, Q2, m2WCET, RBS2, b1, Q1, m1WCET, RBS1, eWCET); 
			}
			else 
				System.out.println("In neither contention nor regulation strictly domiant case but none of the two possible (b1 <= 1/noOfCores && b2 > 1/noOfCores) or (b2 <= 1/noOfCores && b1 > 1/noOfCores) exists.");
		}


		stall += (Constants.period - Math.min(Q1, Q2)); // add the first regulationperiod  
		//System.out.println(stall); 
		return stall ; 

	}

	private long CaseTwoStall_old (long Q1, long cm1_asterisk, long m1WCET, double RBS1, long m2WCET, double RBS2, long eWCET, long WCET)
	{
		long stall = 0; 
		double rm = (double)m2WCET/(double)m1WCET; 
		long cm2_asterisk = (long)Math.max(rm * (double)cm1_asterisk, Math.min(RBS2, ((double)Constants.period/(double)noOfCores)- RBS1)); 


		long n1_asterisk = (long)Math.floor((double)m2WCET/(double)cm2_asterisk); 
		long n2_asterisk = Math.min(m2WCET % cm2_asterisk,1); 
		long n3_asterisk = (long)Math.ceil((double)(m1WCET - (Math.ceil((double)m2WCET/(double)cm2_asterisk)*cm1_asterisk))/(double)cm1_asterisk); 

		long eWCET_asterisk = (n1_asterisk * (Constants.period - (noOfCores * (cm1_asterisk + cm2_asterisk))) ) + 
				(n2_asterisk * (Constants.period - (noOfCores * (cm1_asterisk + (m2WCET % cm2_asterisk)))) )+
				(Math.max(0, n3_asterisk -1) * (Constants.period - (noOfCores * cm1_asterisk)) );

		if(eWCET >= eWCET_asterisk)
		{
			stall = (m1WCET + m2WCET) * (noOfCores -1); 
		}
		else if(eWCET < eWCET_asterisk)
		{
			long c_asterisk = Constants.period - ((noOfCores - 1) * (cm1_asterisk + cm2_asterisk)); 

			if( (long)Math.ceil((double)m1WCET/(double)cm1_asterisk) == (long)Math.ceil((double)m2WCET/(double)cm2_asterisk))
			{
				stall =(long) ((Math.floor((double)WCET/(double)c_asterisk)* (noOfCores-1) * (cm1_asterisk + cm2_asterisk)) + ((noOfCores - 1) * Math.min(WCET % c_asterisk, cm1_asterisk + cm2_asterisk)));   
			}
			else if( (long)Math.ceil((double)m1WCET/(double)cm1_asterisk) > (long)Math.ceil((double)m2WCET/(double)cm2_asterisk))
			{
				long C_asterisk_1 = n1_asterisk * (Constants.period - ((noOfCores -1)*(cm1_asterisk + cm2_asterisk))); 
				long C_asterisk_2 = C_asterisk_1 +	(n2_asterisk * (Constants.period - ((noOfCores -1)*(cm1_asterisk + (m2WCET % cm2_asterisk)))));  

				if(WCET >= C_asterisk_2)
				{
					stall = (n1_asterisk * (cm1_asterisk + cm2_asterisk) * (noOfCores -1)) + (n2_asterisk * (cm1_asterisk + (m2WCET % cm2_asterisk))*(noOfCores -1))
							+ YaoStall(WCET - C_asterisk_2 - Math.min(WCET - C_asterisk_2, m1WCET - ((n1_asterisk + n2_asterisk)*cm1_asterisk)) , Math.min(WCET - C_asterisk_2, m1WCET - ((n1_asterisk + n2_asterisk)*cm1_asterisk)), Q1); 
				}
				else if (WCET >= C_asterisk_1)
				{
					long c2 = Math.min(WCET - C_asterisk_1, Math.min(cm1_asterisk + (m2WCET % cm2_asterisk), (m1WCET + m2WCET) - (n1_asterisk * (cm1_asterisk + cm2_asterisk))) ); 
					stall = (noOfCores - 1) * ((n1_asterisk * (cm1_asterisk + cm2_asterisk)) + c2); 
				}
				else 
				{
					stall = (noOfCores -1) * ( ((long)Math.floor((double)WCET/(double)c_asterisk ) * (cm1_asterisk + cm2_asterisk)) + Math.min(WCET % c_asterisk, cm1_asterisk+cm2_asterisk) ); 
				}

			}
		}

		return stall; 
	}


	private long CaseThreeStall_old(double b1, long Q1,long m1WCET, double RBS1, double b2, long Q2, long m2WCET, double RBS2, long eWCET) 
	{

		long stall = 0; 

		long C1ml = m1WCET % Q1; 

		long eWCET_asterisk = (long)(Math.floor((double)m2WCET/RBS2)* (Constants.period - (noOfCores * RBS2))) - ( C1ml * noOfCores); 
		if(eWCET >= eWCET_asterisk)
		{
			stall = ((long)Math.floor((double)m1WCET/(double)Q1) * (Constants.period - Q1)) + ( (C1ml + m2WCET) * (noOfCores-1) ); 
		}
		else if(eWCET < eWCET_asterisk)
		{
			long deltaCc = Q1 * noOfCores; 
			long deltanc2Max = (long)((RBS2 * (long)Math.floor((double)deltaCc/(double)Q2)) + Math.min(RBS2, deltaCc % Q2)); 
			long deltanc2Min = (long)((RBS2 * (long)Math.floor((double)Math.max(0, deltaCc - (Constants.period - (RBS2*noOfCores)) )/(double)Q2)) + Math.min(RBS2, Math.max(0, deltaCc - (Constants.period - (RBS2*noOfCores))) % Q2));  

			long deltanc2_asterisk = (long)Math.floor((double)(Constants.period - (noOfCores * Q1))/(double)(noOfCores - 1)); 		

			long deltaRho = ComputeRho_old(Q1, RBS1, m1WCET, Q2, RBS2, m2WCET, eWCET, deltanc2Min, deltanc2Max, deltanc2_asterisk, C1ml); 

			stall = YaoStall(0, m1WCET - C1ml - (deltaRho* Q1) , Q1) +
					YaoStall(eWCET + ((C1ml + (deltaRho * Q1)) * noOfCores), m2WCET, Q2); 
		}
		else 
			System.out.println("Error in CaseThreeStall() function as eWCET is neither greather and equal to eWCET_asterisk or less than eWCET_asterisk"); 

		return stall;
	}

	private long ComputeRho_old(long Q1, double RBS1, long m1WCET, long Q2, double RBS2, long m2WCET, long eWCET, long deltanc2Min, long deltanc2Max, long deltanc2_asterisk, long C1ml) 
	{
		long deltarho = 0; 

		if(deltanc2_asterisk >= deltanc2Max)
		{
			deltarho =0; 
		}
		else if (deltanc2_asterisk < deltanc2Min)
		{
			long Ccm2 = 0, Ccm2_temp = 0; 
			long deltarho_temp = 0; 
			long stall = 0, R = 0, Cm1_cap = 0; 

			stall = YaoStall(eWCET + ((C1ml + (deltarho_temp * Q1)) * noOfCores), m2WCET, Q2); 
			R = m2WCET + eWCET + (C1ml*noOfCores) + stall;
			Ccm2 = m2WCET - (long)(Math.floor((double)R/(double)Constants.period)*RBS2) - (long)Math.min( Math.floor( (double)(R % Constants.period) / (double)noOfCores), RBS2); 

			while(Ccm2 > deltanc2_asterisk && deltarho < (long)Math.floor((double)m1WCET/(double)Q1))
			{

				deltarho_temp = deltarho + 1; 
				// QPA style optimisation: deltarho_temp = Math.min(deltarho + (long)Math.floor((double)Ccm2/(double)deltanc2Max), (long)Math.floor((double)m1WCET/(double)Q1)); 

				Cm1_cap = (m1WCET % Q1 ) + (deltarho_temp * Q1) ; 
				stall = YaoStall(eWCET + (Cm1_cap * noOfCores), m2WCET, Q2); 
				R = m2WCET + eWCET + (Cm1_cap * noOfCores) + stall;
				Ccm2_temp = Math.max(0, (m2WCET - (long)(Math.floor((double)R/(double)Constants.period)*RBS2) - (long)Math.min( Math.floor((double)(R % Constants.period)/(double)noOfCores), RBS2) )); 

				if( (Cm1_cap - Math.min(Q1-1, Math.max(0, (long)Math.floor((double)(R%Constants.period)/(double)noOfCores)-RBS2)) )  <  (long)( (Q1-1) * Math.floor((double)R/(double)Constants.period)) )
				{
					deltarho = deltarho_temp; 
					Ccm2 = Ccm2_temp; 
				}
				else 
				{
					break; 
				}

			}

		}
		else if (deltanc2_asterisk >= deltanc2Min && deltanc2_asterisk < deltanc2Max)
		{
			long deltarho_max = 0, stall_max = 0, stall = 0, R = 0, Cm1_cap = 0; 

			//stall = YaoStall(eWCET + (Cm1_cap * noOfCores), m2WCET, Q2); 
			//R = m2WCET + eWCET + (Cm1_cap * noOfCores) + stall;
			//Ccm2_temp = Math.max(0, (m2WCET - (long)(Math.floor((double)R/(double)Constants.period)*RBS2) - (long)Math.min( Math.floor((double)(R % Constants.period)/(double)noOfCores), RBS2) )); 


			for(deltarho = 0; deltarho <= (long)Math.floor((double)m1WCET/(double)Q1); deltarho++)
			{
				Cm1_cap = (m1WCET % Q1 ) + (deltarho * Q1) ; 

				stall = YaoStall(eWCET + (Cm1_cap * noOfCores), m2WCET, Q2); 
				R = m2WCET + eWCET + (Cm1_cap * noOfCores) + stall;

				if ((stall +  ((long)(Math.floor((double)m1WCET/(double)Q1) - deltarho) * (Constants.period - Q1))) > stall_max && 
						(Cm1_cap - Math.min(Q1-1, Math.max(0, (long)Math.floor((double)(R%Constants.period)/(double)noOfCores)-RBS2))) < (long)((Q1 - 1) *  Math.floor((double)R/(double)Constants.period)) )
				{
					stall_max = stall + ((long)(Math.floor((double)m1WCET/(double)Q1) - deltarho) * (Constants.period - Q1)) ;  
					deltarho_max = deltarho; 
				}
			}

			deltarho = deltarho_max; 			
		}
		else 
		{
			System.out.println("deltanc2_asterisk has a strange value and it does not fall in any of the cases in computerho.\t"+deltanc2_asterisk+"\t"+deltanc2Max + "\t" +deltanc2Min);
		}

		return deltarho;
	}


	private long YaoStall(long eWCET, long mWCET, long givenMemoryBudget)
	{
		if(mWCET == 0) // no memory accesses no stall; 
			return 0; 

		long stall = 0; 
		long WCET = eWCET + mWCET; 
		double bandwidth = (double)givenMemoryBudget/(double)Constants.period; 

		if (bandwidth <= (double)1/(double)noOfCores)
		{
			if(mWCET % givenMemoryBudget == 0)
			{
				stall = ((long)((double)mWCET/(double)givenMemoryBudget) * (Constants.period - givenMemoryBudget)) 
						+ (Constants.period - givenMemoryBudget); // sligthy different from Yao which considers only (m-1)Q for last period in this case but we replace it with P-Q
			}
			else 
			{
				stall = ((long)Math.ceil((double)mWCET/(double)givenMemoryBudget) * (Constants.period - givenMemoryBudget)) 
						+ ((noOfCores-1)* (mWCET % givenMemoryBudget)); 
			}
		}
		else
		{

			if(((double)mWCET/(double)WCET) <= (((double)1-bandwidth)/(bandwidth*((double)noOfCores-(double)1))))
			{
				stall = (Constants.period - givenMemoryBudget) + (mWCET * (noOfCores -1));  
			}
			else 
			{
				double RBS = (double)(Constants.period - givenMemoryBudget)/(double)(noOfCores - 1); 
				long K1 = (long)Math.floor((double)eWCET/((double)givenMemoryBudget - RBS)); 

				if(WCET <= (1+K1)*givenMemoryBudget)
				{
					stall = ((1+K1)*(Constants.period-givenMemoryBudget)) +	Math.min(Constants.period - givenMemoryBudget, (long)((noOfCores-1)*(mWCET - K1 * RBS))); 
				}
				else 
				{
					stall = (long)((1+Math.floor((double)WCET/(double)givenMemoryBudget))*(Constants.period-givenMemoryBudget)) +	Math.min(Constants.period - givenMemoryBudget, (noOfCores-1)*(WCET % givenMemoryBudget)); 
				}
			}

		}

		return stall - (Constants.period - givenMemoryBudget); 

	}


	/*private long YaoStall_original(long eWCET, long mWCET, long givenMemoryBudget)
	{
		long stall = 0; 
		long WCET = eWCET + mWCET; 
		double bandwidth = (double)givenMemoryBudget/(double)Constants.period; 

		if (bandwidth <= (double)1/(double)noOfCores)
		{
			if(mWCET % givenMemoryBudget == 0)
			{
				stall = ((long)((double)mWCET/(double)givenMemoryBudget) * (Constants.period - givenMemoryBudget)) 
						+ ((noOfCores-1)*givenMemoryBudget); 
			}
			else 
			{
				stall = ((long)Math.ceil((double)mWCET/(double)givenMemoryBudget) * (Constants.period - givenMemoryBudget)) 
						+ ((noOfCores-1)* (mWCET % givenMemoryBudget)); 
			}
		}
		else
		{

			if(((double)mWCET/(double)WCET) <= (((double)1-bandwidth)/(bandwidth*((double)noOfCores-(double)1))))
			{
				stall = (Constants.period - givenMemoryBudget) + (mWCET * (noOfCores -1));  
			}
			else 
			{
				double RBS = (double)(Constants.period - givenMemoryBudget)/(double)(noOfCores - 1); 
				long K1 = (long)Math.floor((double)eWCET/((double)givenMemoryBudget - RBS)); 

				if(WCET <= (1+K1)*givenMemoryBudget)
				{
					stall = ((1+K1)*(Constants.period-givenMemoryBudget)) +	Math.min(Constants.period - givenMemoryBudget, (long)((noOfCores-1)*(mWCET - K1 * RBS))); 
				}
				else 
				{
					stall = (long)((1+Math.floor((double)WCET/(double)givenMemoryBudget))*(Constants.period-givenMemoryBudget)) +	Math.min(Constants.period - givenMemoryBudget, (noOfCores-1)*(WCET % givenMemoryBudget)); 
				}
			}

		}

		return stall; 

	}*/



}