package Analysis;
import java.util.ArrayList;
import java.util.Collections;

import Common.Constants;
import Common.Enums;
import Common.Task;
import Common.Taskset;


public class Multicontroller_EvenUneven 
{
	private ArrayList<Task> taskSet; 
	private int numberOfCores; 
	private ArrayList<ArrayList<Task>> partitionedTaskset;

	private Enums.ActionResult schedulableEven; 
	private Enums.ActionResult schedulableUnEven;

	private ArrayList<Integer> memoryBudgetOfEachCoreContrOne;
	private ArrayList<Integer> memoryBudgetOfEachCoreContrTwo;

	public Multicontroller_EvenUneven(Taskset ts, int noOfProcessors, int sorting_option)
	{
		numberOfCores = noOfProcessors;
		partitionedTaskset = new ArrayList<ArrayList<Task>>();

		for(int i=0; i<numberOfCores; i++)
		{
			partitionedTaskset.add(new ArrayList<Task>()); // add cores
		}

		schedulableEven = Enums.ActionResult.Failure; 
		schedulableUnEven = Enums.ActionResult.Failure;

		memoryBudgetOfEachCoreContrOne = new ArrayList<Integer>(numberOfCores);
		memoryBudgetOfEachCoreContrTwo = new ArrayList<Integer>(numberOfCores);

		taskSet = new ArrayList<Task>(ts.GetSize()); 
		Task task = ts.GetHeadOfTheList(); 

		while(task!=null)
		{
			taskSet.add(task); 
			task = task.GetNextTask(); 
		}

		for(int i=0; i<numberOfCores; i++)
		{
			memoryBudgetOfEachCoreContrOne.add((int)((double)Constants.totalMemoryBudget/(double)numberOfCores)); 
			memoryBudgetOfEachCoreContrTwo.add((int)((double)Constants.totalMemoryBudget/(double)numberOfCores)); 
		}

		/********** Sort the given array of tasks ****************/
		if(sorting_option == 0) // sort tasks in descending order with repespect to deadlines  
		{
			Collections.sort(taskSet,Task.deadlineComparator);
		}
		else if(sorting_option == 1) // sort tasks in descending order with respect to utilisations 
		{
			Collections.sort(taskSet,Task.utilisationComparator);
		}
		else if(sorting_option == 2) // sort tasks in descending order with repespect to total memory
		{
			Collections.sort(taskSet,Task.totalMemoryComparator);
		}
		else if(sorting_option == 3) // sort tasks in descending order with repespect to total memory density
		{
			Collections.sort(taskSet,Task.totalMemorydensityComparator);
		}
		/*		else if(sorting_option == 4) // sort tasks in descending order with repespect to deadline and then total memory density
		{
			Collections.sort(taskSet,Task.deadlineComparator);
			Collections.sort(taskSet,Task.totalMemorydensityComparator);
		}
		else if(sorting_option == 5) // sorted tasks in desceinding order with respect to utilisation and then total memory density  
		{
			Collections.sort(taskSet,Task.utilisationComparator);
			Collections.sort(taskSet,Task.totalMemorydensityComparator);
		}*/
		/*********************************************************/


		FirstFitPackingAndSchedulabilityTest(); 

	}

	private void FirstFitPackingAndSchedulabilityTest() 
	{

		schedulableEven = Enums.ActionResult.Failure;
		schedulableUnEven = Enums.ActionResult.Failure;
		ArrayList<Task> unassigned = new ArrayList<Task>(); 

		// First fit approach with even distribution of bandwidth
		for(Task task: taskSet)		
		{
			for(int i= 0; i<numberOfCores; i++)
			{
				partitionedTaskset.get(i).add(task); 

				if(new MulticontrollerAnalysis(partitionedTaskset.get(i), memoryBudgetOfEachCoreContrOne.get(i), memoryBudgetOfEachCoreContrTwo.get(i), numberOfCores).GetSchedulabilityStatus() == Enums.ActionResult.Failure)
				{
					partitionedTaskset.get(i).remove(task);
					if(i == numberOfCores-1 || partitionedTaskset.get(i).size() == 0) // the second condition states we cannot assign with even bandwidth
					{
						unassigned.add(task);
						break; 
					}
				}
				else 
				{
					break;
				}
			}
		}

		if(unassigned.size() == 0)
		{
			schedulableEven = Enums.ActionResult.Success;
			schedulableUnEven = Enums.ActionResult.Success;
			return;
		}

		// First fit approach with Uneven bandwidth distribution to assign unallocated tasks
		int reclaimedMemoryBudgetContrOne = 0, currentMemoryBudgetContrOne = 0; 
		int reclaimedMemoryBudgetContrTwo = 0, currentMemoryBudgetContrTwo = 0;
		int[] budgetArray = new int[2]; 

		ArrayList<Task> ts; 
		for(int i=0; i<numberOfCores; i++)
		{
			ts = partitionedTaskset.get(i); 
			
			budgetArray = PerformSensitivityAnalysisToGetTheMinimumBandwidthForEachController(ts,memoryBudgetOfEachCoreContrOne.get(i),memoryBudgetOfEachCoreContrTwo.get(i));
			currentMemoryBudgetContrOne = budgetArray[0]; 
			currentMemoryBudgetContrTwo = budgetArray[1];

			reclaimedMemoryBudgetContrOne += (memoryBudgetOfEachCoreContrOne.get(i)- currentMemoryBudgetContrOne); 
			reclaimedMemoryBudgetContrTwo += (memoryBudgetOfEachCoreContrTwo.get(i)- currentMemoryBudgetContrTwo); 

			memoryBudgetOfEachCoreContrOne.set(i,currentMemoryBudgetContrOne);
			memoryBudgetOfEachCoreContrTwo.set(i,currentMemoryBudgetContrTwo);

		}

		for(Task task: unassigned)		
		{
			for(int i = 0; i<numberOfCores; i++)
			{
				partitionedTaskset.get(i).add(task);

				if((memoryBudgetOfEachCoreContrOne.get(i)+reclaimedMemoryBudgetContrOne > 0 || memoryBudgetOfEachCoreContrTwo.get(i)+reclaimedMemoryBudgetContrTwo > 0) && new MulticontrollerAnalysis(partitionedTaskset.get(i), memoryBudgetOfEachCoreContrOne.get(i)+reclaimedMemoryBudgetContrOne,memoryBudgetOfEachCoreContrTwo.get(i)+reclaimedMemoryBudgetContrTwo, numberOfCores).GetSchedulabilityStatus() == Enums.ActionResult.Success)
				{

					budgetArray = PerformSensitivityAnalysisToGetTheMinimumBandwidthForEachController(partitionedTaskset.get(i), memoryBudgetOfEachCoreContrOne.get(i)+reclaimedMemoryBudgetContrOne, memoryBudgetOfEachCoreContrTwo.get(i)+reclaimedMemoryBudgetContrTwo);
					currentMemoryBudgetContrOne = budgetArray[0]; 
					currentMemoryBudgetContrTwo = budgetArray[1];

					if(currentMemoryBudgetContrOne - memoryBudgetOfEachCoreContrOne.get(i) > 0)
						reclaimedMemoryBudgetContrOne -= (currentMemoryBudgetContrOne - memoryBudgetOfEachCoreContrOne.get(i));
					else 
						reclaimedMemoryBudgetContrOne += (memoryBudgetOfEachCoreContrOne.get(i) - currentMemoryBudgetContrOne);

					if(currentMemoryBudgetContrTwo - memoryBudgetOfEachCoreContrTwo.get(i) > 0) // subtract if we have use more than previous
						reclaimedMemoryBudgetContrTwo -= (currentMemoryBudgetContrTwo - memoryBudgetOfEachCoreContrTwo.get(i));
					else 
						reclaimedMemoryBudgetContrTwo += (memoryBudgetOfEachCoreContrTwo.get(i) - currentMemoryBudgetContrTwo); // add if you are using less than before

					if(currentMemoryBudgetContrOne - memoryBudgetOfEachCoreContrOne.get(i)<0 || currentMemoryBudgetContrTwo - memoryBudgetOfEachCoreContrTwo.get(i)<0) // just a test to see if it is possible 
					{
						System.out.println("New bandwidth (with additional) task required is less than the previous" + currentMemoryBudgetContrOne +"\t" + memoryBudgetOfEachCoreContrOne.get(i)+ "\t" +currentMemoryBudgetContrTwo +"\t" + memoryBudgetOfEachCoreContrTwo.get(i));
					}

					memoryBudgetOfEachCoreContrOne.set(i, currentMemoryBudgetContrOne); 
					memoryBudgetOfEachCoreContrTwo.set(i, currentMemoryBudgetContrTwo); 

					break; 
				}
				else 
				{
					if(i==numberOfCores-1)
					{
						schedulableEven = Enums.ActionResult.Failure;
						schedulableUnEven = Enums.ActionResult.Failure;
						return; 
					}
					else 
					{
						partitionedTaskset.get(i).remove(task); 
					}

				}
			}
		}

		schedulableEven = Enums.ActionResult.Failure;
		schedulableUnEven = Enums.ActionResult.Success;

		//System.out.print(partitionedTaskset.size()+"\t");
	}

	private int[] PerformSensitivityAnalysisToGetTheMinimumBandwidthForEachController(ArrayList<Task> ts, int current_MemoryBandwidthContrOne, int current_MemoryBandwidthContrTwo) // assumes that the task-set is feasible with current_MemoryBandwidth
	{
		if(ts.isEmpty())
			return new int[]{0,0} ; 

		int X = 0; 
		Enums.ActionResult status = null; 

		int firstContrOne = 0, secondContrOne = current_MemoryBandwidthContrOne;
		int firstContrTwo = 0, secondContrTwo = current_MemoryBandwidthContrTwo;
		int diff = Constants.MemoryBandwidthQuantization; 


		while(secondContrOne - firstContrOne > diff || secondContrTwo - firstContrTwo > diff )
		{
			if(secondContrOne - firstContrOne > diff)
			{
				X = (int)(Math.floor(((double)firstContrOne + (double)secondContrOne)/2)); 
				status = new MulticontrollerAnalysis(ts, X, secondContrTwo, numberOfCores).GetSchedulabilityStatus(); 
				if( status == Enums.ActionResult.Success)
					secondContrOne = X; 
				else 
					firstContrOne = X;  
			}

			if(secondContrTwo - firstContrTwo > diff)
			{
				X = (int)(Math.floor(((double)firstContrTwo + (double)secondContrTwo)/2)); 
				status = new MulticontrollerAnalysis(ts, secondContrOne, X, numberOfCores).GetSchedulabilityStatus(); 
				if( status == Enums.ActionResult.Success)
					secondContrTwo = X; 
				else 
					firstContrTwo = X;  
			}

		}

		//System.out.print("Old:\t"+secondContrOne +"\t"+ secondContrTwo + "\t sum:\t" + (secondContrOne+secondContrTwo)+ "\tNew:\t"); 
		return new int[]{secondContrOne, secondContrTwo} ;
	}


/*	private int[] PerformSensitivityAnalysisToGetTheMinimumBandwidthForEachController(ArrayList<Task> ts, int current_MemoryBandwidthContrOne, int current_MemoryBandwidthContrTwo) // assumes that the task-set is feasible with current_MemoryBandwidth
	{
		if(ts.isEmpty())
			return new int[]{0,0} ; 

		int X = 0; 
		Enums.ActionResult status = null; 

		int firstContrOne = 0, secondContrOne = current_MemoryBandwidthContrOne;
		int firstContrTwo = 0, secondContrTwo = current_MemoryBandwidthContrTwo;
		int diff = Constants.MemoryBandwidthQuantization; 

		while(secondContrOne - firstContrOne > diff )
		{
			X = (int)(Math.floor(((double)firstContrOne + (double)secondContrOne)/2)); 
			status = new MulticontrollerAnalysis(ts, X, secondContrTwo, numberOfCores).GetSchedulabilityStatus(); 
			if( status == Enums.ActionResult.Success)
				secondContrOne = X; 
			else 
				firstContrOne = X;  

		}

		while(secondContrTwo - firstContrTwo > diff )
		{	

			X = (int)(Math.floor(((double)firstContrTwo + (double)secondContrTwo)/2)); 
			status = new MulticontrollerAnalysis(ts, secondContrOne, X, numberOfCores).GetSchedulabilityStatus(); 
			if( status == Enums.ActionResult.Success)
				secondContrTwo = X; 
			else 
				firstContrTwo = X;  

		}
		
		int [] order_first_second = new int[]{secondContrOne, secondContrTwo}; 
		
		
		
		X = 0; 
		status = null; 
		firstContrOne = 0; 
		secondContrOne = current_MemoryBandwidthContrOne;
		firstContrTwo = 0; 
		secondContrTwo = current_MemoryBandwidthContrTwo;
		diff = Constants.MemoryBandwidthQuantization; 

		while(secondContrTwo - firstContrTwo > diff )
		{	

			X = (int)(Math.floor(((double)firstContrTwo + (double)secondContrTwo)/2)); 
			status = new MulticontrollerAnalysis(ts, secondContrOne, X, numberOfCores).GetSchedulabilityStatus(); 
			if( status == Enums.ActionResult.Success)
				secondContrTwo = X; 
			else 
				firstContrTwo = X;  
		}
		
		while(secondContrOne - firstContrOne > diff )
		{
			X = (int)(Math.floor(((double)firstContrOne + (double)secondContrOne)/2)); 
			status = new MulticontrollerAnalysis(ts, X, secondContrTwo, numberOfCores).GetSchedulabilityStatus(); 
			if( status == Enums.ActionResult.Success)
				secondContrOne = X; 
			else 
				firstContrOne = X;  
		}


		int[] order_second_first = new int[]{secondContrOne, secondContrTwo}; 	

		
		//System.out.println(order_first_second[0]+ "\t"+order_first_second[1] +"\t sum: \t"+ (order_first_second[0]+order_first_second[1] )+"\t"+order_second_first[0]+ "\t"+order_second_first[1] +"\t sum: \t"+ (order_second_first[0]+order_second_first[1] ) + "\t" + (order_first_second[0]+order_first_second[1] - order_second_first[0] -order_second_first[1]) );
		
		if(order_first_second[0]+order_first_second[1] < order_second_first[0]+order_second_first[1] )
			return order_first_second;
		else 
			return order_second_first;
	}
*/
	public Enums.ActionResult GetEvenSchedulabilityStatus()
	{
		return schedulableEven; 
	}

	public Enums.ActionResult GetUnEvenSchedulabilityStatus()
	{
		return schedulableUnEven; 
	}

	public int GetNumberOfProcessorsUsed()
	{
		return partitionedTaskset.size(); 
	}

	public ArrayList<Integer> GetBandwidthAllocationContrOne()
	{
		return memoryBudgetOfEachCoreContrOne; 
	}

	public ArrayList<Integer> GetBandwidthAllocationContrTwo()
	{
		return memoryBudgetOfEachCoreContrTwo; 
	}

	/*public int GetBandwidthAllocationForGivenCoreForGivenController(int core, int Controller)
	{
		if(Controller == 1)
			return memoryBudgetOfEachCoreContrOne.get(core);
		else if(Controller == 2)
			return memoryBudgetOfEachCoreContrTwo.get(core);
		else
		{
			System.out.println("Wrong Controller number");
			return -1; 
		}
	}*/
}
