package Analysis;
import java.util.ArrayList;
import java.util.Collections;

import Common.Constants;
import Common.Enums;
import Common.Task;
import Common.Taskset;

public class Multicontroller_Humblefit 
{
	private ArrayList<Task> taskSet; 
	private int numberOfCores; 
	private ArrayList<ArrayList<Task>> partitionedTaskset;

	private Enums.ActionResult schedulable;

	private ArrayList<Integer> memoryBudgetOfEachCoreContrOne; 
	private ArrayList<Integer> memoryBudgetOfEachCoreContrTwo; 


	public Multicontroller_Humblefit(Taskset ts, int noOfProcessors, int sorting_option)
	{
		numberOfCores = noOfProcessors;
		partitionedTaskset = new ArrayList<ArrayList<Task>>();
		for(int i=0; i<numberOfCores; i++)
		{
			partitionedTaskset.add(new ArrayList<Task>()); // add cores
		}

		schedulable= Enums.ActionResult.Failure; 

		memoryBudgetOfEachCoreContrOne = new ArrayList<Integer>(numberOfCores);
		memoryBudgetOfEachCoreContrTwo = new ArrayList<Integer>(numberOfCores);


		taskSet = new ArrayList<Task>(ts.GetSize()); 
		Task task = ts.GetHeadOfTheList(); 

		while(task!=null)
		{
			taskSet.add(task); 
			task = task.GetNextTask(); 
		}

		for(int i=0; i<numberOfCores; i++)
		{
			memoryBudgetOfEachCoreContrOne.add(0); 
			memoryBudgetOfEachCoreContrTwo.add(0); 
		}
		/********** Sort the given array of tasks ****************/
		if(sorting_option == 0) // sort tasks in descending order with repespect to deadlines  
		{
			Collections.sort(taskSet,Task.deadlineComparator);
		}
		else if(sorting_option == 1) // sort tasks in descending order with respect to utilisations 
		{
			Collections.sort(taskSet,Task.utilisationComparator);
		}
		else if(sorting_option == 2) // sort tasks in descending order with repespect to total memory
		{
			Collections.sort(taskSet,Task.totalMemoryComparator);
		}
		else if(sorting_option == 3) // sort tasks in descending order with repespect to total memory density
		{
			Collections.sort(taskSet,Task.totalMemorydensityComparator);
		}
		/*		else if(sorting_option == 4) // sort tasks in descending order with repespect to deadline and then total memory density
		{
			Collections.sort(taskSet,Task.deadlineComparator);
			Collections.sort(taskSet,Task.totalMemorydensityComparator);
		}
		else if(sorting_option == 5) // sorted tasks in desceinding order with respect to utilisation and then total memory density  
		{
			Collections.sort(taskSet,Task.utilisationComparator);
			Collections.sort(taskSet,Task.totalMemorydensityComparator);
		}*/
		/*********************************************************/

		FirstFitPackingAndSchedulabilityTest(); 

	}

	private void FirstFitPackingAndSchedulabilityTest() 
	{

		schedulable = Enums.ActionResult.Failure;
		int [] assigned = new int[taskSet.size()];  
		int [] budgetArray = new int[]{0,0}; 

		int remainingMemoryBudgetContrOne =  Constants.totalMemoryBudget, requiredMemoryBudgetContrOne = 0;
		int remainingMemoryBudgetContrTwo =  Constants.totalMemoryBudget, requiredMemoryBudgetContrTwo = 0;

		// First fit approach with all bandwidth allocated to first core. We check all tasks on each core without backtracking
		for(int i= 0; i<numberOfCores; i++)
		{
			memoryBudgetOfEachCoreContrOne.set(i, remainingMemoryBudgetContrOne); 
			memoryBudgetOfEachCoreContrTwo.set(i, remainingMemoryBudgetContrTwo); 
			
			for(Task task: taskSet)		
			{
				if(assigned[taskSet.indexOf(task)]==0)
				{
					partitionedTaskset.get(i).add(task); 
					if((remainingMemoryBudgetContrOne <= 0 && task.GetM_WCET(0)>0) || (remainingMemoryBudgetContrTwo <= 0 && task.GetM_WCET(1) > 0) || new MulticontrollerAnalysis(partitionedTaskset.get(i), memoryBudgetOfEachCoreContrOne.get(i),memoryBudgetOfEachCoreContrTwo.get(i), numberOfCores).GetSchedulabilityStatus() == Enums.ActionResult.Failure)
					{
						partitionedTaskset.get(i).remove(task);
						if(i == numberOfCores-1 || partitionedTaskset.get(i).size() == 0) // the second condition states we cannot assign with the remaining bandwidth
							return; 
						
						break; // The task not feasible in this core. try the second one 
					}
					else 
					{
						assigned[taskSet.indexOf(task)] = 1;
					}
				}
			}

			budgetArray = PerformSensitivityAnalysisToGetTheMinimumBandwidthForEachController(partitionedTaskset.get(i),memoryBudgetOfEachCoreContrOne.get(i), memoryBudgetOfEachCoreContrTwo.get(i));
			requiredMemoryBudgetContrOne = budgetArray[0]; 
			requiredMemoryBudgetContrTwo = budgetArray[1]; 
			
			if(requiredMemoryBudgetContrOne > memoryBudgetOfEachCoreContrOne.get(i) || requiredMemoryBudgetContrTwo > memoryBudgetOfEachCoreContrTwo.get(i))
			{
				System.out.println("The requireBandwidth is greater than the already assigned bandwidth \t"+ requiredMemoryBudgetContrOne + "\t"+memoryBudgetOfEachCoreContrOne.get(i)+ "\t"+ requiredMemoryBudgetContrTwo + "\t"+memoryBudgetOfEachCoreContrTwo.get(i));
			}
			
			remainingMemoryBudgetContrOne = memoryBudgetOfEachCoreContrOne.get(i) - requiredMemoryBudgetContrOne;
			remainingMemoryBudgetContrTwo = memoryBudgetOfEachCoreContrTwo.get(i) - requiredMemoryBudgetContrTwo; 

			memoryBudgetOfEachCoreContrOne.set(i,requiredMemoryBudgetContrOne);
			memoryBudgetOfEachCoreContrTwo.set(i,requiredMemoryBudgetContrTwo);

		}
		schedulable = Enums.ActionResult.Success;

		//System.out.print(partitionedTaskset.size()+"\t");
	}

	private int[] PerformSensitivityAnalysisToGetTheMinimumBandwidthForEachController(ArrayList<Task> ts, int current_MemoryBandwidthContrOne, int current_MemoryBandwidthContrTwo) // assumes that the task-set is feasible with current_MemoryBandwidth
	{
		if(ts.isEmpty())
			return new int[]{0,0} ; 

		int X = 0; 
		Enums.ActionResult status = null; 

		int firstContrOne = 0, secondContrOne = current_MemoryBandwidthContrOne;
		int firstContrTwo = 0, secondContrTwo = current_MemoryBandwidthContrTwo;
		int diff = Constants.MemoryBandwidthQuantization; 


		while(secondContrOne - firstContrOne > diff || secondContrTwo - firstContrTwo > diff )
		{
			if(secondContrOne - firstContrOne > diff)
			{
				X = (int)(Math.floor(((double)firstContrOne + (double)secondContrOne)/2)); 
				status = new MulticontrollerAnalysis(ts, X, secondContrTwo, numberOfCores).GetSchedulabilityStatus(); 
				if( status == Enums.ActionResult.Success)
					secondContrOne = X; 
				else 
					firstContrOne = X;  
			}
			
			if(secondContrTwo - firstContrTwo > diff)
			{
				X = (int)(Math.floor(((double)firstContrTwo + (double)secondContrTwo)/2)); 
				status = new MulticontrollerAnalysis(ts, secondContrOne, X, numberOfCores).GetSchedulabilityStatus(); 
				if( status == Enums.ActionResult.Success)
					secondContrTwo = X; 
				else 
					firstContrTwo = X;  
			}

		}

		return new int[]{secondContrOne, secondContrTwo} ;
	}
	
/*	private int[] PerformSensitivityAnalysisToGetTheMinimumBandwidthForEachController(ArrayList<Task> ts, int current_MemoryBandwidthContrOne, int current_MemoryBandwidthContrTwo) // assumes that the task-set is feasible with current_MemoryBandwidth
	{
		if(ts.isEmpty())
			return new int[]{0,0} ; 

		int X = 0; 
		Enums.ActionResult status = null; 

		int firstContrOne = 0, secondContrOne = current_MemoryBandwidthContrOne;
		int firstContrTwo = 0, secondContrTwo = current_MemoryBandwidthContrTwo;
		int diff = Constants.MemoryBandwidthQuantization; 

		while(secondContrOne - firstContrOne > diff )
		{
			X = (int)(Math.floor(((double)firstContrOne + (double)secondContrOne)/2)); 
			status = new MulticontrollerAnalysis(ts, X, secondContrTwo, numberOfCores).GetSchedulabilityStatus(); 
			if( status == Enums.ActionResult.Success)
				secondContrOne = X; 
			else 
				firstContrOne = X;  

		}

		while(secondContrTwo - firstContrTwo > diff )
		{	

			X = (int)(Math.floor(((double)firstContrTwo + (double)secondContrTwo)/2)); 
			status = new MulticontrollerAnalysis(ts, secondContrOne, X, numberOfCores).GetSchedulabilityStatus(); 
			if( status == Enums.ActionResult.Success)
				secondContrTwo = X; 
			else 
				firstContrTwo = X;  

		}
		
		int [] order_first_second = new int[]{secondContrOne, secondContrTwo}; 
		
		
		
		X = 0; 
		status = null; 
		firstContrOne = 0; 
		secondContrOne = current_MemoryBandwidthContrOne;
		firstContrTwo = 0; 
		secondContrTwo = current_MemoryBandwidthContrTwo;
		diff = Constants.MemoryBandwidthQuantization; 

		while(secondContrTwo - firstContrTwo > diff )
		{	

			X = (int)(Math.floor(((double)firstContrTwo + (double)secondContrTwo)/2)); 
			status = new MulticontrollerAnalysis(ts, secondContrOne, X, numberOfCores).GetSchedulabilityStatus(); 
			if( status == Enums.ActionResult.Success)
				secondContrTwo = X; 
			else 
				firstContrTwo = X;  
		}
		
		while(secondContrOne - firstContrOne > diff )
		{
			X = (int)(Math.floor(((double)firstContrOne + (double)secondContrOne)/2)); 
			status = new MulticontrollerAnalysis(ts, X, secondContrTwo, numberOfCores).GetSchedulabilityStatus(); 
			if( status == Enums.ActionResult.Success)
				secondContrOne = X; 
			else 
				firstContrOne = X;  
		}


		int[] order_second_first = new int[]{secondContrOne, secondContrTwo}; 	

		
		//System.out.println(order_first_second[0]+ "\t"+order_first_second[1] +"\t"+ order_second_first[0]+ "\t"+order_second_first[1] + "\t" + (order_first_second[0]+order_first_second[1] - order_second_first[0] -order_second_first[1]) );
		
		if(order_first_second[0]+order_first_second[1] < order_second_first[0]+order_second_first[1] )
			return order_first_second;
		else 
			return order_second_first;
	}
*/

	public Enums.ActionResult GetSchedulabilityStatus()
	{
		return schedulable; 
	}

	public int GetNumberOfProcessorsUsed()
	{
		return partitionedTaskset.size(); 
	}

	public ArrayList<Integer> GetBandwidthAllocationContrOne()
	{
		return memoryBudgetOfEachCoreContrOne; 
	}
	
	public ArrayList<Integer> GetBandwidthAllocationContrTwo()
	{
		return memoryBudgetOfEachCoreContrTwo; 
	}

	/*public int GetBandwidthAllocationForGivenCoreForGivenController(int core, int Controller)
	{
		if(Controller == 1)
			return memoryBudgetOfEachCoreContrOne.get(core);
		else if(Controller == 2)
			return memoryBudgetOfEachCoreContrTwo.get(core);
		else
		{
			System.out.println("Wrong Controller number");
			return -1; 
		}
	}*/
}
