package Analysis;
import java.util.ArrayList;
import java.util.Collections;

import Common.Constants;
import Common.Enums;
import Common.Task;
import Common.Taskset;
import Common.Utils;

public class PartitionedYao 
{
	private ArrayList<Task> taskSet; 
	private int numberOfCores; 
	private ArrayList<ArrayList<Task>> partitionedTaskset;

	private Enums.ActionResult schedulableEven;
	private Enums.ActionResult schedulableUnEven;

	private ArrayList<Integer> memoryBudgetOfEachCore; 


	public PartitionedYao(Taskset ts, int noOfProcessors, int sorting_option)
	{
		numberOfCores = noOfProcessors;
		partitionedTaskset = new ArrayList<ArrayList<Task>>();

		for(int i=0; i<numberOfCores; i++)
		{
			partitionedTaskset.add(new ArrayList<Task>()); // add list for each cores
		}

		schedulableEven = Enums.ActionResult.Failure; 
		schedulableUnEven = Enums.ActionResult.Failure; 

		memoryBudgetOfEachCore = new ArrayList<Integer>(numberOfCores);

		taskSet = new ArrayList<Task>(ts.GetSize()); 
		Task task = ts.GetHeadOfTheList(); 

		while(task!=null)
		{
			taskSet.add(task); 
			task = task.GetNextTask(); 
		}

		for(int i=0; i<numberOfCores; i++)
		{
			memoryBudgetOfEachCore.add((int)((double)Constants.totalMemoryBudget/(double)(numberOfCores/2))); // initialize with even value
		}

		/********** Sort the given array of tasks ****************/
		if(sorting_option == 0) // sort tasks in descending order with repespect to deadlines  
		{
			Collections.sort(taskSet,Task.deadlineComparator);
		}
		else if(sorting_option == 1) // sort tasks in descending order with respect to utilisations 
		{
			Collections.sort(taskSet,Task.utilisationComparator);
		}
		else if(sorting_option == 2) // sort tasks in descending order with repespect to total memory
		{
			Collections.sort(taskSet,Task.totalMemoryComparator);
		}
		else if(sorting_option == 3) // sort tasks in descending order with repespect to total memory density
		{
			Collections.sort(taskSet,Task.totalMemorydensityComparator);
		}
		/*		else if(sorting_option == 4) // sort tasks in descending order with repespect to deadline and then total memory density
		{
			Collections.sort(taskSet,Task.deadlineComparator);
			Collections.sort(taskSet,Task.totalMemorydensityComparator);
		}
		else if(sorting_option == 5) // sorted tasks in desceinding order with respect to utilisation and then total memory density  
		{
			Collections.sort(taskSet,Task.utilisationComparator);
			Collections.sort(taskSet,Task.totalMemorydensityComparator);
		}*/
		/*********************************************************/

		FirstFitPackingAndSchedulabilityTest(); 

	}

	private void FirstFitPackingAndSchedulabilityTest() 
	{


		schedulableEven = Enums.ActionResult.Failure;
		schedulableUnEven = Enums.ActionResult.Failure;

		ArrayList<Task> unassigned = new ArrayList<Task>(); 


		// First fit approach with even distribution of bandwidth
		for(Task task: taskSet)		
		{
			for(int i= 0; i<numberOfCores; i++)
			{
				partitionedTaskset.get(i).add(task); 

				if(new YaoAnalysis(partitionedTaskset.get(i), memoryBudgetOfEachCore.get(i), numberOfCores/2).GetSchedulabilityStatus() == Enums.ActionResult.Failure)
				{
					partitionedTaskset.get(i).remove(task);
					if(i == numberOfCores-1 || partitionedTaskset.get(i).size() == 0) // the second condition states we cannot assign with even bandwidth
					{
						unassigned.add(task);
						break; 
					}
				}
				else 
				{
					break;
				}
			}
		}

		if(unassigned.size() == 0)
		{
			schedulableEven = Enums.ActionResult.Success;
			schedulableUnEven = Enums.ActionResult.Success;
			return;
		}

		
		// First fit approach with Uneven bandwidth distribution to assign unallocated tasks to first set of cores 
		
		ArrayList<Task> unassignedRemaining = new ArrayList<Task>(); 
		int reclaimedMemoryBudget = 0, currentMemoryBudget = 0; 
		ArrayList<Task> ts; 
		for(int i=0; i<numberOfCores/2; i++)
		{
			ts = partitionedTaskset.get(i); 
			currentMemoryBudget = PerformSensitivityAnalysisToGetTheMinimumBandwidth(ts,memoryBudgetOfEachCore.get(i), numberOfCores/2); 
			reclaimedMemoryBudget += memoryBudgetOfEachCore.get(i)- currentMemoryBudget; 
			memoryBudgetOfEachCore.set(i,currentMemoryBudget);
		}

		for(Task task: unassigned)		
		{
			for(int i= 0; i < numberOfCores/2; i++)
			{
				partitionedTaskset.get(i).add(task);

				if(memoryBudgetOfEachCore.get(i)+reclaimedMemoryBudget>0 && new YaoAnalysis(partitionedTaskset.get(i), memoryBudgetOfEachCore.get(i)+reclaimedMemoryBudget, numberOfCores/2).GetSchedulabilityStatus() == Enums.ActionResult.Success)
				{
					currentMemoryBudget = PerformSensitivityAnalysisToGetTheMinimumBandwidth(partitionedTaskset.get(i), memoryBudgetOfEachCore.get(i)+reclaimedMemoryBudget, numberOfCores/2);
					reclaimedMemoryBudget -= (currentMemoryBudget - memoryBudgetOfEachCore.get(i)); 

					if(currentMemoryBudget - memoryBudgetOfEachCore.get(i)<0) // just a test to see if it is possible 
					{
						System.out.println("New bandwidth (with additional) task required is less than the previous" + currentMemoryBudget +"\t" + memoryBudgetOfEachCore.get(i));
					}

					memoryBudgetOfEachCore.set(i, currentMemoryBudget); 
					
					
					break; 
				}
				else 
				{
					if(i==(numberOfCores/2)-1)
					{
						unassignedRemaining.add(task);  
					}
					else 
					{
						partitionedTaskset.get(i).remove(task); 
					}				
				}
			}
		}



		// First fit approach with Uneven bandwidth distribution to assign unallocated tasks to SECOND set of cores 
		reclaimedMemoryBudget = 0; 
		currentMemoryBudget = 0; 
		for(int i=numberOfCores/2; i<numberOfCores; i++)
		{
			ts = partitionedTaskset.get(i); 
			currentMemoryBudget = PerformSensitivityAnalysisToGetTheMinimumBandwidth(ts,memoryBudgetOfEachCore.get(i), numberOfCores/2); 
			reclaimedMemoryBudget += memoryBudgetOfEachCore.get(i)- currentMemoryBudget; 
			memoryBudgetOfEachCore.set(i,currentMemoryBudget);
		}

		for(Task task: unassignedRemaining)		
		{
			for(int i= numberOfCores/2; i < numberOfCores; i++)
			{
				partitionedTaskset.get(i).add(task);

				if(memoryBudgetOfEachCore.get(i)+reclaimedMemoryBudget>0 && new YaoAnalysis(partitionedTaskset.get(i), memoryBudgetOfEachCore.get(i)+reclaimedMemoryBudget, numberOfCores/2).GetSchedulabilityStatus() == Enums.ActionResult.Success)
				{
					currentMemoryBudget = PerformSensitivityAnalysisToGetTheMinimumBandwidth(partitionedTaskset.get(i), memoryBudgetOfEachCore.get(i)+reclaimedMemoryBudget, numberOfCores/2);
					reclaimedMemoryBudget -= (currentMemoryBudget - memoryBudgetOfEachCore.get(i)); 

					if(currentMemoryBudget - memoryBudgetOfEachCore.get(i)<0) // just a test to see if it is possible 
					{
						System.out.println("New bandwidth (with additional) task required is less than the previous" + currentMemoryBudget +"\t" + memoryBudgetOfEachCore.get(i));
					}

					memoryBudgetOfEachCore.set(i, currentMemoryBudget); 
					break; 
				}
				else 
				{
					if(i==numberOfCores-1)
					{
						schedulableEven = Enums.ActionResult.Failure;
						schedulableUnEven = Enums.ActionResult.Failure;
						return; 
					}
					else 
					{
						partitionedTaskset.get(i).remove(task); 
					}

				}
			}
		}

		schedulableEven = Enums.ActionResult.Failure;
		schedulableUnEven = Enums.ActionResult.Success;

		//System.out.print(partitionedTaskset.size()+"\t");
	}

	private int PerformSensitivityAnalysisToGetTheMinimumBandwidth(ArrayList<Task> ts, int current_MemoryBandwidth, int cores)
	{
		if(ts.isEmpty())
			return 0; 

		int X = 0; 
		int first = 0, second = current_MemoryBandwidth; 
		int diff = Constants.MemoryBandwidthQuantization; 
		Enums.ActionResult status = null; 

		while(second - first > diff)
		{
			X = (int)(Math.floor(((double)second + (double)first)/2)); 
			status = new YaoAnalysis(ts, X, cores).GetSchedulabilityStatus(); 
			if( status == Enums.ActionResult.Success)
			{
				second = Utils.ConvertDoubleToInt(Math.floor(((double)second + (double)first)/2));
			}
			else 
			{
				first = Utils.ConvertDoubleToInt(Math.floor(((double)second + (double)first)/2)); 
			}
		}

		return second;
	}


	public Enums.ActionResult GetEvenSchedulabilityStatus()
	{
		return schedulableEven; 
	}

	public Enums.ActionResult GetUnEvenSchedulabilityStatus()
	{
		return schedulableUnEven; 
	}

	public int GetNumberOfProcessorsUsed()
	{
		return partitionedTaskset.size(); 
	}

	public ArrayList<Integer> GetBandwidthAllocation()
	{
		return memoryBudgetOfEachCore; 
	}

	public int GetBandwidthAllocationForGivenCore(int core)
	{
		return memoryBudgetOfEachCore.get(core); 
	}
}
