package Analysis;
import java.util.ArrayList;
import java.util.Collections;

import Common.Constants;
import Common.Enums;
import Common.Task;
import Common.Taskset;
import Common.Utils;

public class PartitionedYao_Greedyfit 
{
	private ArrayList<Task> taskSet; 
	private int numberOfCores; 
	private ArrayList<ArrayList<Task>> partitionedTaskset;

	private Enums.ActionResult schedulable;

	private ArrayList<Integer> memoryBudgetOfEachCore; 


	public PartitionedYao_Greedyfit(Taskset ts, int noOfProcessors, int sorting_option)
	{
		numberOfCores = noOfProcessors;
		partitionedTaskset = new ArrayList<ArrayList<Task>>();
		for(int i=0; i<numberOfCores; i++)
		{
			partitionedTaskset.add(new ArrayList<Task>()); // add cores

		}

		schedulable= Enums.ActionResult.Failure; 

		memoryBudgetOfEachCore = new ArrayList<Integer>(numberOfCores);

		taskSet = new ArrayList<Task>(ts.GetSize()); 
		Task task = ts.GetHeadOfTheList(); 

		while(task!=null)
		{
			taskSet.add(task); 
			task = task.GetNextTask(); 
		}

		for(int i=0; i<numberOfCores; i++)
		{
			memoryBudgetOfEachCore.add(0); 
		}
		/********** Sort the given array of tasks ****************/
		if(sorting_option == 0) // sort tasks in descending order with repespect to deadlines  
		{
			Collections.sort(taskSet,Task.deadlineComparator);
		}
		else if(sorting_option == 1) // sort tasks in descending order with respect to utilisations 
		{
			Collections.sort(taskSet,Task.utilisationComparator);
		}
		else if(sorting_option == 2) // sort tasks in descending order with repespect to total memory
		{
			Collections.sort(taskSet,Task.totalMemoryComparator);
		}
		else if(sorting_option == 3) // sort tasks in descending order with repespect to total memory density
		{
			Collections.sort(taskSet,Task.totalMemorydensityComparator);
		}
		/*		else if(sorting_option == 4) // sort tasks in descending order with repespect to deadline and then total memory density
		{
			Collections.sort(taskSet,Task.deadlineComparator);
			Collections.sort(taskSet,Task.totalMemorydensityComparator);
		}
		else if(sorting_option == 5) // sorted tasks in desceinding order with respect to utilisation and then total memory density  
		{
			Collections.sort(taskSet,Task.utilisationComparator);
			Collections.sort(taskSet,Task.totalMemorydensityComparator);
		}*/
		/*********************************************************/

		FirstFitPackingAndSchedulabilityTest(); 

	}

	private void FirstFitPackingAndSchedulabilityTest() 
	{

		schedulable = Enums.ActionResult.Failure;
		int [] assigned = new int[taskSet.size()];  

		
		int remainingMemoryBudgetSetOne =  Constants.totalMemoryBudget, requiredMemoryBudget = 0;
		int remainingMemoryBudgetSetTwo =  Constants.totalMemoryBudget;

		int remainingMemoryBudget = 0;
		

		// First fit approach with all bandwidth allocated to first core. We check all tasks on each core without backtracking
		for(int i= 0; i<numberOfCores; i++)
		{
			
			remainingMemoryBudget = i<numberOfCores/2 ? remainingMemoryBudgetSetOne: remainingMemoryBudgetSetTwo;
			
			memoryBudgetOfEachCore.set(i, remainingMemoryBudget); 

			for(Task task: taskSet)		
			{
				if(assigned[taskSet.indexOf(task)]==0)
				{
					partitionedTaskset.get(i).add(task); 
					if(remainingMemoryBudget <= 0 || new YaoAnalysis(partitionedTaskset.get(i), memoryBudgetOfEachCore.get(i), numberOfCores/2).GetSchedulabilityStatus() == Enums.ActionResult.Failure)
					{
						partitionedTaskset.get(i).remove(task);
						if(i == numberOfCores-1 || (partitionedTaskset.get(i).size() == 0 && (i>=numberOfCores/2))) // the second condition states we cannot assign with the remaining bandwidth
							return; 
					}
					else 
					{
						assigned[taskSet.indexOf(task)] = 1;
					}
				}
			}

			requiredMemoryBudget = PerformSensitivityAnalysisToGetTheMinimumBandwidth(partitionedTaskset.get(i),memoryBudgetOfEachCore.get(i), numberOfCores/2);
			if(requiredMemoryBudget > memoryBudgetOfEachCore.get(i))
			{
				System.out.println("The requireBandwidth is greater than the already assigned bandwidth \t"+ requiredMemoryBudget + "\t"+memoryBudgetOfEachCore.get(i));
			}

			if(i<numberOfCores/2)
				remainingMemoryBudgetSetOne = memoryBudgetOfEachCore.get(i) - requiredMemoryBudget;
			else 
				remainingMemoryBudgetSetTwo = memoryBudgetOfEachCore.get(i) - requiredMemoryBudget;

			memoryBudgetOfEachCore.set(i,requiredMemoryBudget);

		}
		schedulable = Enums.ActionResult.Success;

		//System.out.print(partitionedTaskset.size()+"\t");
	}

	private int PerformSensitivityAnalysisToGetTheMinimumBandwidth(ArrayList<Task> ts, int current_MemoryBandwidth, int cores)
	{
		if(ts.isEmpty())
			return 0; 

		int X = 0; 
		int first = 0, second = current_MemoryBandwidth; 
		int diff = Constants.MemoryBandwidthQuantization; 
		Enums.ActionResult status = null; 

		while(second - first > diff)
		{
			X = (int)(Math.floor(((double)second + (double)first)/2)); 
			status = new YaoAnalysis(ts, X, cores).GetSchedulabilityStatus(); 
			if( status == Enums.ActionResult.Success)
			{
				second = Utils.ConvertDoubleToInt(Math.floor(((double)second + (double)first)/2));
			}
			else 
			{
				first = Utils.ConvertDoubleToInt(Math.floor(((double)second + (double)first)/2)); 
			}
		}

		return second;
	}


	
	public Enums.ActionResult GetSchedulabilityStatus()
	{
		return schedulable; 
	}

	public int GetNumberOfProcessorsUsed()
	{
		return partitionedTaskset.size(); 
	}

	public ArrayList<Integer> GetBandwidthAllocation()
	{
		return memoryBudgetOfEachCore; 
	}

	public int GetBandwidthAllocationForGivenCore(int core)
	{
		return memoryBudgetOfEachCore.get(core); 
	}
}
