package Analysis;
import java.util.ArrayList;
import java.util.Collections;

import Common.Constants;
import Common.Enums;
import Common.Task;
import Common.Taskset;
import Common.Utils;

public class PartitionedYao_Memoryfit 
{
	private ArrayList<Task> taskSet; 
	private int numberOfCores; 
	private ArrayList<ArrayList<Task>> partitionedTaskset;

	private Enums.ActionResult schedulable;

	private ArrayList<Integer> memoryBudgetOfEachCore; 

	
	public PartitionedYao_Memoryfit(Taskset ts, int noOfProcessors, int sorting_option)
	{
		numberOfCores = noOfProcessors;
		partitionedTaskset = new ArrayList<ArrayList<Task>>();

		for(int i=0; i<numberOfCores; i++)
		{
			partitionedTaskset.add(new ArrayList<Task>()); // add cores
		}

		schedulable = Enums.ActionResult.Failure; 

		memoryBudgetOfEachCore = new ArrayList<Integer>(numberOfCores);

		taskSet = new ArrayList<Task>(ts.GetSize()); 
		Task task = ts.GetHeadOfTheList(); 

		while(task!=null)
		{
			taskSet.add(task); 
			task = task.GetNextTask(); 
		}

		for(int i=0; i<numberOfCores; i++)
		{
			memoryBudgetOfEachCore.add(0); 
		}

		/********** Sort the given array of tasks ****************/
		if(sorting_option == 0) // sort tasks in descending order with repespect to deadlines  
		{
			Collections.sort(taskSet,Task.deadlineComparator);
		}
		else if(sorting_option == 1) // sort tasks in descending order with respect to utilisations 
		{
			Collections.sort(taskSet,Task.utilisationComparator);
		}
		else if(sorting_option == 2) // sort tasks in descending order with repespect to total memory
		{
			Collections.sort(taskSet,Task.totalMemoryComparator);
		}
		else if(sorting_option == 3) // sort tasks in descending order with repespect to total memory density
		{
			Collections.sort(taskSet,Task.totalMemorydensityComparator);
		}
		/*		else if(sorting_option == 4) // sort tasks in descending order with repespect to deadline and then total memory density
		{
			Collections.sort(taskSet,Task.deadlineComparator);
			Collections.sort(taskSet,Task.totalMemorydensityComparator);
		}
		else if(sorting_option == 5) // sorted tasks in desceinding order with respect to utilisation and then total memory density  
		{
			Collections.sort(taskSet,Task.utilisationComparator);
			Collections.sort(taskSet,Task.totalMemorydensityComparator);
		}*/
		/*********************************************************/

		FirstFitPackingAndSchedulabilityTest(); 

	}

	private void FirstFitPackingAndSchedulabilityTest() 
	{

		schedulable = Enums.ActionResult.Failure;
		int[] deltaIncrease = new int[numberOfCores]; 
		int index_increase =0, increase = 0; 

		int remainingMemoryBudgetSetOne =  Constants.totalMemoryBudget;
		int remainingMemoryBudgetSetTwo =  Constants.totalMemoryBudget;
		int remainingMemoryBudget = 0; 

		// First fit approach with memorycentric distribution of bandwidth
		for(Task task: taskSet)		
		{
			for(int i= 0; i<numberOfCores; i++)
			{
				
				partitionedTaskset.get(i).add(task); 

				remainingMemoryBudget = i<numberOfCores/2 ? remainingMemoryBudgetSetOne : remainingMemoryBudgetSetTwo; 
				if(remainingMemoryBudget <= 0 || new YaoAnalysis(partitionedTaskset.get(i), memoryBudgetOfEachCore.get(i)+remainingMemoryBudget, numberOfCores/2).GetSchedulabilityStatus() == Enums.ActionResult.Failure)
					deltaIncrease[i] = -1; 
				else
					deltaIncrease[i] = PerformSensitivityAnalysisToGetTheMinimumBandwidth(partitionedTaskset.get(i), memoryBudgetOfEachCore.get(i)+remainingMemoryBudget, numberOfCores/2) 
										- memoryBudgetOfEachCore.get(i); 
								
				partitionedTaskset.get(i).remove(task);
				
				if(i == numberOfCores -1)
				{
					index_increase = Integer.MAX_VALUE;
					increase = Integer.MAX_VALUE; 
					for(int j=0; j<deltaIncrease.length; j++)
					{
						if(deltaIncrease[j]!=-1 && (deltaIncrease[j] < increase))
						{
							index_increase = j; 
							increase = deltaIncrease[j]; 
						}
					}
					
					if(index_increase != Integer.MAX_VALUE && increase != Integer.MAX_VALUE)
					{
						partitionedTaskset.get(index_increase).add(task); 
						memoryBudgetOfEachCore.set(index_increase, memoryBudgetOfEachCore.get(index_increase)+deltaIncrease[index_increase]); 
						
						if(index_increase<numberOfCores/2)
							remainingMemoryBudgetSetOne -= deltaIncrease[index_increase];
						else 
							remainingMemoryBudgetSetTwo -= deltaIncrease[index_increase];
					}
					else 
						return; 
					
				}
				
			}
		}
		
		schedulable = Enums.ActionResult.Success;

		//System.out.print(partitionedTaskset.size()+"\t");
	}

	private int PerformSensitivityAnalysisToGetTheMinimumBandwidth(ArrayList<Task> ts, int current_MemoryBandwidth, int cores)
	{
		if(ts.isEmpty())
			return 0; 

		int X = 0; 
		int first = 0, second = current_MemoryBandwidth; 
		int diff = Constants.MemoryBandwidthQuantization; 
		Enums.ActionResult status = null; 

		while(second - first > diff)
		{
			X = (int)(Math.floor(((double)second + (double)first)/2)); 
			status = new YaoAnalysis(ts, X, cores).GetSchedulabilityStatus(); 
			if( status == Enums.ActionResult.Success)
			{
				second = Utils.ConvertDoubleToInt(Math.floor(((double)second + (double)first)/2));
			}
			else 
			{
				first = Utils.ConvertDoubleToInt(Math.floor(((double)second + (double)first)/2)); 
			}
		}

		return second;
	}



	public Enums.ActionResult GetSchedulabilityStatus()
	{
		return schedulable; 
	}

		public int GetNumberOfProcessorsUsed()
	{
		return partitionedTaskset.size(); 
	}

	public ArrayList<Integer> GetBandwidthAllocation()
	{
		return memoryBudgetOfEachCore; 
	}

	public int GetBandwidthAllocationForGivenCore(int core)
	{
		return memoryBudgetOfEachCore.get(core); 
	}
}
