package Analysis;
import java.util.ArrayList;
import Common.Constants;
import Common.Enums;
import Common.Task;


public class YaoAnalysis 
{
	private ArrayList<Task> taskset;
	private Enums.ActionResult schedulable; 
	private long noOfCores; 
	private long givenMemoryBudget; 

	public YaoAnalysis(ArrayList<Task> ts, long memorybudget, long cores)
	{		
		noOfCores = cores; 
		givenMemoryBudget = memorybudget; 
		taskset = new ArrayList<Task>(ts.size()); 
		taskset = ts; 
		schedulable = Enums.ActionResult.Failure; 
		schedulable = AssignPriorities(); 
	}


	private Enums.ActionResult AssignPriorities() 
	{

		boolean flag = false; 

		for(Task task:taskset)
			task.SetPriority(-1);

		for(int k=taskset.size()-1; k>=0; k--)
		{
			flag = false; 
			for(Task task:taskset)
			{
				if(task.GetPriority() == -1)
				{
					task.SetPriority(k);
					if(PerformYaoAnalysis(task) == Enums.ActionResult.Failure)
					{
						task.SetPriority(-1);
					}
					else
					{
						flag = true; 
						break; 
					}
				}
			}

			if(flag == false)
				return Enums.ActionResult.Failure; 
		}

		return Enums.ActionResult.Success; 
	}


	private Enums.ActionResult PerformYaoAnalysis(Task task) 
	{

		long responseTimeWithoutStall = 0;  
		long responseTimeWithStall = 0;  

		responseTimeWithoutStall = ComputeResponseTime(task, 0, 0); 
		if(responseTimeWithoutStall > (long)task.GetDeadline())
			return Enums.ActionResult.Failure;



		responseTimeWithStall = ComputeResponseTime(task, responseTimeWithoutStall, 1); 
		if(responseTimeWithStall > (long)task.GetDeadline() )
			return Enums.ActionResult.Failure;

		// if all the tests are passed then pass success
		return Enums.ActionResult.Success; 

	}

	private long ComputeResponseTime(Task task, long startResponseTime, int stall) 
	{
		long prevResponseTime =-1, currResponseTime = startResponseTime; 
		long [] response = {0,0,0};  

		while(currResponseTime != prevResponseTime && currResponseTime <=task.GetDeadline())
		{
			prevResponseTime = currResponseTime;
			response = ComputeDemand(currResponseTime,task); 
			currResponseTime = (long)task.GetWCET() + response[0];

			if(stall == 1) //Response time with stall
				currResponseTime += Stall(response[1],response[2],response[1]+response[2]);
		}

		return currResponseTime;
	}


	private long[] ComputeDemand(long time, Task task) 
	{
		long[] totaldemand = {0,0,0}; 

		long jobs = 0; 

		for(Task t: taskset) 
		{
			jobs = (long) Math.ceil((double)time/(double)t.GetMinInterArrival()); 

			if(t.GetPriority()<=task.GetPriority())
			{
				if(!task.equals(t))
					totaldemand[0] += (jobs * (long)t.GetWCET()); 
			
				totaldemand[1] += (jobs * (long)t.GetE_WCET()); 
				totaldemand[2] += (jobs * (long)t.GetTotalM_WCET()); 
			}

		}

		return totaldemand;
	}

	private long Stall(long eWCET, long mWCET, long WCET)
	{
		long stall = 0; 
		double bandwidth = (double)givenMemoryBudget/(double)Constants.period; 

		if (bandwidth <= (double)1/(double)noOfCores)
		{
			if(mWCET % givenMemoryBudget == 0)
			{
				stall = ((long)((double)mWCET/(double)givenMemoryBudget) * (Constants.period - givenMemoryBudget)) 
						+ ((noOfCores-1)*givenMemoryBudget); 
			}
			else 
			{
				stall = ((long)Math.ceil((double)mWCET/(double)givenMemoryBudget) * (Constants.period - givenMemoryBudget)) 
						+ ((noOfCores-1)* (mWCET % givenMemoryBudget)); 
			}
		}
		else
		{

			if(((double)mWCET/(double)WCET) <= (((double)1-bandwidth)/(bandwidth*((double)noOfCores-(double)1))))
			{
				stall = (Constants.period - givenMemoryBudget) + (mWCET * (noOfCores -1));  
			}
			else 
			{
				double RBS = (double)(Constants.period - givenMemoryBudget)/(double)(noOfCores - 1); 
				long K1 = (long)Math.floor((double)eWCET/((double)givenMemoryBudget - RBS)); 

				if(WCET <= (1+K1)*givenMemoryBudget)
				{
					stall = ((1+K1)*(Constants.period-givenMemoryBudget)) +	Math.min(Constants.period - givenMemoryBudget, (long)((noOfCores-1)*(mWCET - K1 * RBS))); 
				}
				else 
				{
					stall = (long)((1+Math.floor((double)WCET/(double)givenMemoryBudget))*(Constants.period-givenMemoryBudget)) +	Math.min(Constants.period - givenMemoryBudget, (noOfCores-1)*(WCET % givenMemoryBudget)); 
				}
			}

		}

		return stall; 

	}

	public Enums.ActionResult GetSchedulabilityStatus()
	{
		return schedulable; 
	}

	public int GetMaximumDeadline() 
	{

		int maximumdeadline = 0; 		
		for(Task task: taskset)
		{
			if(maximumdeadline<task.GetDeadline())
				maximumdeadline = task.GetDeadline(); 
		}

		return maximumdeadline;
	}


}