package Common;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;

import Analysis.Multicontroller_EvenUneven;
import Analysis.Multicontroller_Greedyfit;
import Analysis.Multicontroller_Humblefit;
import Analysis.Multicontroller_Memoryfit;
import Analysis.PartitionedYao_EvenUneven;
import Analysis.PartitionedYao_Greedyfit;
import Analysis.PartitionedYao_Humblefit;
import Analysis.PartitionedYao_Memoryfit;


public class Main 
{
	public static void main(String[] args) throws IOException 
	{
		AbstractParameters(args); 
	}

	public static void AbstractParameters(String[] args) throws IOException 
	{

		int arg_index=0; 

		if( args.length != 14 && args.length != 17 ) {
			System.out.println("Usage: java Main <noOfProcessors> <numberOfTasks>" 
					+ "<systemUtilization (0.3)>" 
					+ "<regulationPeriod (1,000,000)> <totalMemoryBudget (1,000,000)>"
					+ "<sorting_options (4 by memory density)>" 
					+ "<memory intensity (0: SPEC benchmark)>"	// otherwise maximum C^m/T
					+ "<MemoryBandwidthQuantization (1)>"
					+ "<start_repetitions (0)> <end_repetitions (1000)>"
					+ "<deadlineModel (0: D=T)>"
					+ "<minInterArrivalTimeLimit(10,000)> <maxInterArrivalTimeLimit (1,000,000)>" // range of task period
					+ "<results_file_name>");	
			return;
		}

				int noOfProcessors 		 = Utils.ConvertStringToInt(args[arg_index++].replace(',', '.'));
		int numberOfTasks 		 = Utils.ConvertStringToInt(args[arg_index++].replace(',', '.'));
		double systemUtilization = Utils.ConvertStringToDouble(args[arg_index++].replace(',', '.'));

		//Constants.L = Utils.ConvertStringToDouble(args[arg_index++]);
		Constants.period = Utils.ConvertStringToInt(args[arg_index++].replace(',', '.'));
		Constants.totalMemoryBudget =  Utils.ConvertStringToInt(args[arg_index++].replace(',', '.')); //(int) ((double)Constants.period/Constants.L); 

		int sorting_option = Utils.ConvertStringToInt(args[arg_index++].replace(',', '.'));
		double memory_intensity = Utils.ConvertStringToDouble(args[arg_index++]);

		// fixed in script 
		Constants.MemoryBandwidthQuantization =  Utils.ConvertStringToInt(args[arg_index++].replace(',', '.')); 
		int start_repetitions	 	= Utils.ConvertStringToInt(args[arg_index++].replace(',', '.')); 
		int end_repetitions 	 	= Utils.ConvertStringToInt(args[arg_index++].replace(',', '.'));
		int deadlineModel 		 = Utils.ConvertStringToInt(args[arg_index++].replace(',', '.'));
		int minInterArrivalTimeLimit = Utils.ConvertStringToInt(args[arg_index++].replace(',', '.')); 
		int maxInterArrivalTimeLimit = Utils.ConvertStringToInt(args[arg_index++].replace(',', '.')); 
		String file_name = args[arg_index++].replace(',', '.'); 	

		// fixed parameters
		RandomGeneratorUsingJavaRandomClass RNG_Util = new RandomGeneratorUsingJavaRandomClass(1); 
		RandomGeneratorUsingJavaRandomClass RNG_MinimumArrivalTime = new RandomGeneratorUsingJavaRandomClass(3); 
		RandomGeneratorUsingJavaRandomClass RNG_Deadline = new RandomGeneratorUsingJavaRandomClass(5); 
		RandomGeneratorUsingJavaRandomClass RNG_Memory = new RandomGeneratorUsingJavaRandomClass(7); 
		systemUtilization = systemUtilization * noOfProcessors; 


		FileWriter fstream = new FileWriter(file_name, true);;  
		BufferedWriter out = new BufferedWriter(fstream);

		for(int i=0; i<end_repetitions ; i++)
		{
			
			Taskset taskset = new Taskset();
			taskset.CreateTasksetWithAbstractParameters(numberOfTasks, systemUtilization, noOfProcessors, minInterArrivalTimeLimit, maxInterArrivalTimeLimit, deadlineModel, RNG_Util,RNG_MinimumArrivalTime, RNG_Deadline, RNG_Memory, memory_intensity);
	
			if(i<start_repetitions)
				continue; 

			if(taskset.GetStatus()== Enums.ActionResult.Success)
			{
				

				PartitionedYao_EvenUneven pyao_even_uneven = new PartitionedYao_EvenUneven(taskset, noOfProcessors, sorting_option); 
				PartitionedYao_Greedyfit pyao_greedyfit = new PartitionedYao_Greedyfit(taskset, noOfProcessors, sorting_option);
				PartitionedYao_Humblefit pyao_humblefit = new PartitionedYao_Humblefit(taskset, noOfProcessors, sorting_option); 
				PartitionedYao_Memoryfit pyao_memoryfit = new PartitionedYao_Memoryfit(taskset, noOfProcessors, sorting_option); 


				Multicontroller_EvenUneven mc_even_uneven = new Multicontroller_EvenUneven(taskset, noOfProcessors, sorting_option); 
				Multicontroller_Greedyfit mc_greedyfit = new Multicontroller_Greedyfit(taskset, noOfProcessors, sorting_option);
				Multicontroller_Humblefit mc_humblefit = new Multicontroller_Humblefit(taskset, noOfProcessors, sorting_option); 
				Multicontroller_Memoryfit mc_memoryfit = new Multicontroller_Memoryfit(taskset, noOfProcessors, sorting_option); 

				out.write(i+ "\t" + noOfProcessors + "\t" + numberOfTasks + "\t" + systemUtilization + "\t" + Constants.period + "\t" + Constants.totalMemoryBudget + "\t" + sorting_option
						+ "\t" + memory_intensity + "\t" + Constants.MemoryBandwidthQuantization  + "\t" + start_repetitions + "\t" + end_repetitions + "\t" + deadlineModel+ "\t"+minInterArrivalTimeLimit +
						"\t"+ maxInterArrivalTimeLimit + "\t" + Status(taskset.GetStatus())+"\t"+Status(pyao_even_uneven.GetEvenSchedulabilityStatus())+ "\t" +Status(pyao_even_uneven.GetUnEvenSchedulabilityStatus())+ "\t" +
						Status(pyao_greedyfit.GetSchedulabilityStatus())+ "\t" + Status(pyao_humblefit.GetSchedulabilityStatus())+ "\t" +Status(pyao_memoryfit.GetSchedulabilityStatus())
						+"\t"+Status(mc_even_uneven.GetEvenSchedulabilityStatus())+ "\t" +Status(mc_even_uneven.GetUnEvenSchedulabilityStatus())+ "\t" + 
						Status(mc_greedyfit.GetSchedulabilityStatus())+ "\t" + Status(mc_humblefit.GetSchedulabilityStatus())+ "\t" +Status(mc_memoryfit.GetSchedulabilityStatus())+"\n");  


				/*System.out.println(i+ "\t" + noOfProcessors + "\t" + numberOfTasks + "\t" + systemUtilization + "\t" + Constants.period + "\t" + Constants.totalMemoryBudget + "\t" + sorting_option
						+ "\t" + memory_intensity + "\t" + Constants.MemoryBandwidthQuantization  + "\t" + start_repetitions + "\t" + end_repetitions + "\t" + deadlineModel+ "\t"+minInterArrivalTimeLimit +
						"\t"+ maxInterArrivalTimeLimit + "\t" + Status(taskset.GetStatus())+"\t"+Status(pyao_even_uneven.GetEvenSchedulabilityStatus())+ "\t" +Status(pyao_even_uneven.GetUnEvenSchedulabilityStatus())+ "\t" +
						Status(pyao_greedyfit.GetSchedulabilityStatus())+ "\t" + Status(pyao_humblefit.GetSchedulabilityStatus())+ "\t" +Status(pyao_memoryfit.GetSchedulabilityStatus())
						+"\t"+Status(mc_even_uneven.GetEvenSchedulabilityStatus())+ "\t" +Status(mc_even_uneven.GetUnEvenSchedulabilityStatus())+ "\t" + 
						Status(mc_greedyfit.GetSchedulabilityStatus())+ "\t" + Status(mc_humblefit.GetSchedulabilityStatus())+ "\t" +Status(mc_memoryfit.GetSchedulabilityStatus())+"\n");    
*/				
			}
			else 
			{

				out.write(i+ "\t" + noOfProcessors + "\t" + numberOfTasks + "\t" + systemUtilization + "\t" + Constants.period + "\t" + Constants.totalMemoryBudget + "\t" + sorting_option
						+ "\t" + memory_intensity + "\t" + Constants.MemoryBandwidthQuantization  + "\t" + start_repetitions + "\t" + end_repetitions + "\t" + deadlineModel+ "\t"+minInterArrivalTimeLimit +
						"\t"+ maxInterArrivalTimeLimit + "\t" + Status(taskset.GetStatus())+"\t"+ 0 + "\t" + 0 + "\t" +	0 + "\t" + 0 + "\t" + 0
						+"\t"+ 0 + "\t" + 0 + "\t" +	0 + "\t" + 0 + "\t" + 0 +"\n");

		/*		System.out.println(i+ "\t" + noOfProcessors + "\t" + numberOfTasks + "\t" + systemUtilization + "\t" + Constants.period + "\t" + Constants.totalMemoryBudget + "\t" + sorting_option
						+ "\t" + memory_intensity + "\t" + Constants.MemoryBandwidthQuantization  + "\t" + start_repetitions + "\t" + end_repetitions + "\t" + deadlineModel+ "\t"+minInterArrivalTimeLimit +
						"\t"+ maxInterArrivalTimeLimit + "\t" + Status(taskset.GetStatus())+"\t"+ 0 + "\t" + 0 + "\t" +	0 + "\t" + 0 + "\t" + 0
						+"\t"+ 0 + "\t" + 0 + "\t" +	0 + "\t" + 0 + "\t" + 0 +"\n");  
*/				
			}
		}		
		out.close();
	}



	private static int Status(Enums.ActionResult ar)
	{
		if(ar == Enums.ActionResult.Success)
			return 1; 
		else if(ar == Enums.ActionResult.Failure)
			return 0; 
		else 
		{
			System.out.print("Status of the schedulability is unknown" + ar + "\t");
			return -1; 
		}
	}

	public static void PrintTasksetParamters(Taskset taskset)
	{
		Task ts = taskset.GetHeadOfTheList(); 
		while(ts!=null)
		{
			ts.Write();
			ts = ts.GetNextTask(); 
		}

	}


	public static void PrintSystemParameters(String args[],double systemUtilization)
	{
		for(int i=0;i<args.length;i++)
			System.out.print(args[i]+"\t"); 

		System.out.print(systemUtilization); 
	}

	public static double TotalUtilisation(Taskset ts)
	{
		double totalutils = 0; 
		Task task= ts.GetHeadOfTheList(); 
		while(task!=null)
		{
			totalutils+=task.GetUtilisation(); 
			task = task.GetNextTask(); 
		}

		return totalutils; 
	}


}