package Common;

import java.util.ArrayList;
import java.util.Comparator;

public class Task 
{	
	private static int globalId = 1;
	private int id, WCET, e_WCET, minInterArrival, deadline;
	
	private ArrayList<Integer> m_WCET; // it will have two elements, memory accesses for LWCET and HWCET

	private double utilization; // it will have two element, LUtilisation and HUtilisation
	
	private int priority; 
	private Task nextTask;

	public Task(int mia, int dl, int wcet, ArrayList<Integer> m_wcet, int e_wcet, double u, int pr, Task next)
	{
		id = globalId++;
		minInterArrival = mia;
		deadline = dl;
 		WCET = wcet;
 		
 		SetM_WCETForBothControllers(m_wcet);
 		e_WCET = e_wcet; 
		utilization = u;
		
		priority = pr; 
		nextTask = next;
	}	

	public int GetPriority()
	{
		return priority;
	}
	
	public void SetPriority(int pr)
	{
		priority = pr; 
	}
	
	public int GetID()
	{
		return id;
	}

	public int GetMinInterArrival()
	{
		return minInterArrival;
	}

	public int GetDeadline()
	{
		return deadline; 
	}

	public int GetWCET()
	{
		return WCET; 
	}


	public int GetE_WCET()
	{
		return e_WCET; 
	}


	public int GetM_WCET(int controller)
	{
		return m_WCET.get(controller); 
	}
	
	public int GetTotalM_WCET()
	{
		return m_WCET.get(0)+m_WCET.get(1); 
	}

	private void SetM_WCETForBothControllers(ArrayList<Integer> m_wcet_in) 
	{
		m_WCET = new ArrayList<Integer>(m_wcet_in.size()); 
		m_WCET.addAll(m_wcet_in); 
	}
	
	public double GetUtilisation()
	{
		return utilization; 
	}

	
	public ArrayList<Integer> GetM_WCETVector()
	{
		return m_WCET; 
	}
	
	public Enums.ActionResult SetNextTask(Task next)
	{
		nextTask = next;
		return Enums.ActionResult.Success;
	}

	public Task GetNextTask()
	{
		return nextTask;
	}

	
	public void Write()
	{
		System.out.println("TaskID = " + id + "\t WCET" + WCET + "\t LWCET with interference" + "\t deadline = " + deadline + "\t Min-intrArvl = " + minInterArrival );
	}

	public void Print()
	{
		System.out.println("TaskID = " + id + "\t WCET" + WCET  +"\t deadline = " + deadline + "\t Min-intrArvl = " + minInterArrival );
	}


	

	/*Comparator for sorting the arraylist by deadline*/
	public static Comparator<Task> deadlineComparator = new Comparator<Task>() 
	{

		public int compare(Task t1, Task t2) 
		{
			int deadline_t1 = t1.GetDeadline(); 
			int deadline_t2 = t2.GetDeadline(); 

			//descending order
			return deadline_t2-deadline_t1;
			//ascending order
			//return deadline_t1-deadline_t2;
		}
	};

	/*Comparator for sorting the list by utilisation*/
	public static Comparator<Task> utilisationComparator = new Comparator<Task>() 
	{

		public int compare(Task t1, Task t2) 
		{
			Double utilisation_t1 =  (Double)((double)t1.GetWCET()/(double)t1.GetMinInterArrival()); 
			Double utilisation_t2 = (Double)((double)t2.GetWCET()/(double)t2.GetMinInterArrival()); 

			//For descending order
			return utilisation_t2.compareTo(utilisation_t1);
			//For ascending order
			//return utilisation_t2.compareTo(utilisation_t1); 
			
		}
	};	
	
	/*Comparator for sorting the list by criticality*/
	public static Comparator<Task> totalMemoryComparator = new Comparator<Task>() 
	{

		public int compare(Task t1, Task t2) 
		{
			int memory_t1 = t1.GetM_WCET(0) + t1.GetM_WCET(1);  
			int memory_t2 = t2.GetM_WCET(0) + t2.GetM_WCET(1);  

			//descending order
			// return criticality_t2 - criticality_t1; 
			//ascending order
			return memory_t2 - memory_t1; 
		}
	};
	
	/*Comparator for sorting the list by priority*/
	public static Comparator<Task> priorityComparator = new Comparator<Task>() 
	{

		public int compare(Task t1, Task t2) 
		{
			int priority_t1 = t1.GetPriority();  
			int priority_t2 = t2.GetPriority(); 

			//descending order
			// return criticality_t2 - criticality_t1; 
			//ascending order
			return priority_t2 - priority_t1; 
		}
	};
	
	/*Comparator for sorting the list by priority*/
	public static Comparator<Task> totalMemorydensityComparator = new Comparator<Task>() 
	{

		public int compare(Task t1, Task t2) 
		{
			Double mem_density_t1 = (Double)(((double)t1.GetM_WCET(0)+(double)t1.GetM_WCET(1))/(double)t1.GetMinInterArrival());   
			Double mem_density_t2 = (Double)(((double)t2.GetM_WCET(0)+(double)t2.GetM_WCET(1))/(double)t2.GetMinInterArrival());   

			//For descending order
			return mem_density_t2.compareTo(mem_density_t1); 
			//For ascending order
			//return utilisation_t2.compareTo(utilisation_t1); 
			
		}
	};
	
	/*Comparator for sorting the list by memory density of first controller*/
	public static Comparator<Task> memorydensityControllerOneComparator = new Comparator<Task>() 
	{

		public int compare(Task t1, Task t2) 
		{
			Double mem_density_t1 = (Double)((double)t1.GetM_WCET(0)/(double)t1.GetMinInterArrival());   
			Double mem_density_t2 = (Double)((double)t2.GetM_WCET(0)/(double)t2.GetMinInterArrival());   

			//For descending order
			return mem_density_t2.compareTo(mem_density_t1); 
			//For ascending order
			//return utilisation_t2.compareTo(utilisation_t1); 
			
		}
	};
	
	/*Comparator for sorting the list by memory density of second controller*/
	public static Comparator<Task> memorydensityControllerSecondComparator = new Comparator<Task>() 
	{

		public int compare(Task t1, Task t2) 
		{
			Double mem_density_t1 = (Double)((double)t1.GetM_WCET(1)/(double)t1.GetMinInterArrival());   
			Double mem_density_t2 = (Double)((double)t2.GetM_WCET(1)/(double)t2.GetMinInterArrival());   

			//For descending order
			return mem_density_t2.compareTo(mem_density_t1); 
			//For ascending order
			//return utilisation_t2.compareTo(utilisation_t1); 
			
		}
	};
}


