package Common;

import java.util.ArrayList;

public class Taskset 
{
	private Task headOfTheList;
	private int taskSetSize;
	private double totalUtilization;
	private Enums.ActionResult status; 

	private Taskset(String inputFile)
	{
		headOfTheList = null;
		taskSetSize = 0;
		totalUtilization = 0;
		status = null;

	}

	public Taskset()
	{
		this("");
	}

	public void CreateTasksetWithAbstractParameters(int numberOfTasks, double systemUtilization, int noProcessors, int minInterArrivalTimeLimit, int maxInterArrivalTimeLimit, int deadlineModel, RandomGeneratorUsingJavaRandomClass RNG_util, RandomGeneratorUsingJavaRandomClass RNG_Period, RandomGeneratorUsingJavaRandomClass RNG_Deadline, RandomGeneratorUsingJavaRandomClass RNG_Memory, double memory_intensity)
	{

		double[] util;
		if (systemUtilization<=noProcessors)
			util = GenerateUtilizations(numberOfTasks, systemUtilization, noProcessors, RNG_util);
		else 
		{
			util = GenerateUtilizations(numberOfTasks, noProcessors, noProcessors, RNG_util); 
			for (int i=0; i<util.length; i++) 
			{
				util[i] *= systemUtilization/(double)noProcessors;
			}
		}

		// Initialise constant variables. 
		Constants.TmaxRT = maxInterArrivalTimeLimit; 
		Constants.TminRT = minInterArrivalTimeLimit; 
		Constants.Tmin_lower_bound =  Math.log10(Constants.TminRT);
		Constants.Tmax_upper_bound =  Math.log10(Constants.TmaxRT);

		/***************************************************************************/
		int wcet = 0, e_wcet = 0, minInterArrival = 0, deadline = 0; 
		double utilisation =0; 

		double r_ratio[] = {0, 0, 0.01, 0.01, 0.02, 0.02, 0.02, 0.03, 0.03, 0.06, 0.06, 0.13, 0.17, 0.17, 0.24, 0.26, 0.31, 0.35, 0.35, 0.36, 0.37, 0.47, 0.61};
		int memory_rand, total_memory_accesses, contr_1_accesses, contr_2_accesses; 

		for (int i = 0; i < util.length; i++) 
		{
			minInterArrival = Utils.ConvertDoubleToInt(Math.floor(Math.pow(10, RNG_Period.GetNextDouble(Constants.Tmin_lower_bound, Constants.Tmax_upper_bound))))*Constants.Resolution;  

			wcet = Utils.MaxInt(1, Utils.ConvertDoubleToInt(Utils.ConvertIntToDouble(minInterArrival)*util[i])); 

			if(deadlineModel == 0)
				deadline = minInterArrival; // deadlineMode == 0 means implicit deadlines 
			else if(deadlineModel == 1 && wcet < minInterArrival) 
				deadline = RNG_Deadline.GetNextInt(wcet, minInterArrival+1);  // deadlineMode == 1 means constrained deadline model, moreover +1 is added with minInterarrival time to include it in the limits
			else if(deadlineModel == 1 && wcet == minInterArrival)
				deadline = minInterArrival;
			else 
				System.out.print("WCET is greater than minimum inter-arrival time or deadlineModel > 1. \t ");

			ArrayList<Integer> m_wcet = new ArrayList<Integer>();

			if(memory_intensity == 0)
			{
				memory_rand = RNG_Memory.GetNextInt(0, r_ratio.length);

				if(r_ratio[memory_rand] != 0 )
				{
					total_memory_accesses = (int)((double)wcet * r_ratio[memory_rand]); 

					if(total_memory_accesses !=0)
					{	
						contr_1_accesses = RNG_Memory.GetNextInt(0,total_memory_accesses); 
						contr_2_accesses = total_memory_accesses - contr_1_accesses;
					}
					else 
					{
						contr_1_accesses = 0; 
						contr_2_accesses = 0;
					}
				}
				else 
				{
					total_memory_accesses = 0; 
					contr_1_accesses = 0; 
					contr_2_accesses = 0; 
				}

				m_wcet.add(contr_1_accesses);  
				m_wcet.add(contr_2_accesses);
			}
			else 
			{
				total_memory_accesses = (int)((double)wcet * RNG_Memory.GetNextDouble(0, memory_intensity)); 

				if(total_memory_accesses != 0)
				{
					contr_1_accesses = RNG_Memory.GetNextInt(0,total_memory_accesses); 
					contr_2_accesses = total_memory_accesses - contr_1_accesses; 
				}
				else 
				{
					contr_1_accesses = 0; 
					contr_2_accesses = 0; 
				}

				m_wcet.add(contr_1_accesses);  
				m_wcet.add(contr_2_accesses);
			}

			e_wcet = wcet - m_wcet.get(0) - m_wcet.get(1);   

			utilisation = (double)wcet/(double)minInterArrival; 


			Task task = new Task(minInterArrival, deadline, wcet, m_wcet,e_wcet, utilisation, -1, null); 

			AddTaskInInputOrder(task);//taskset.AddTaskInInputOrder(task); 
		}


		ValidTaskSet(noProcessors); 

	}


	public static int GetPoisson(double lambda, RandomGeneratorUsingJavaRandomClass RNG_Poisson) 
	{
		double L = Math.exp(-lambda);
		double p = 1.0;
		int k = 0;

		do 
		{
			k++;
			p *= RNG_Poisson.GetNextDouble(); 
		} while (p > L);

		return k - 1;
	}


	private double[] GenerateUtilizations(int numberOfTasks, double targetUtilisation, int noProcessors, RandomGeneratorUsingJavaRandomClass RNG_util) 
	{
		double[] util = new double[numberOfTasks];
		double nextSumU;
		boolean discard;

		do {
			double sumU = targetUtilisation;
			discard = false;
			for (int i = 0; i < numberOfTasks - 1; i++) 
			{
				nextSumU = sumU	* Math.pow(RNG_util.GetNextDoubleExcludingZero(), (double) 1	/ (numberOfTasks - (i + 1))); // zero is excluded while generating random number
				util[i] = sumU - nextSumU;
				sumU = nextSumU;
				if (util[i] > 1) 
				{
					discard = true;
				}
			}
			util[numberOfTasks - 1] = sumU;
			if (util[numberOfTasks - 1] > 1) 
			{
				discard = true;
			}
		} while (discard || !utilizationIsValid(util, noProcessors));
		return util;
	}

	private boolean utilizationIsValid(double[] util, int noProcessors) 
	{
		double sum = 0;

		for (double u : util) 
		{
			sum += u;
		}

		return sum <= noProcessors ? true : false;
	}

	private void ValidTaskSet(/*Taskset ts,*/ int noOfProcessors)
	{
		double task_util = 0, total_util =0 ;  
		Enums.ActionResult flag = Enums.ActionResult.Success;
		Task task = GetHeadOfTheList(); 

		while(task!=null)
		{
			task_util = task.GetUtilisation(); 
			total_util += task_util;
			if(task_util > 1)
			{
				flag = Enums.ActionResult.Failure;
				break;
			}

			task = task.GetNextTask();
		}

		if(flag == Enums.ActionResult.Failure || total_util > noOfProcessors)
			status = Enums.ActionResult.Failure;
		else 
			status = Enums.ActionResult.Success;  
	}


	public Enums.ActionResult GetStatus()
	{
		return status;
	}

	public Task GetHeadOfTheList()
	{
		return headOfTheList;
	}

	public double GetTotalUtilization()
	{
		return totalUtilization;
	}

	public int GetSize()
	{
		return taskSetSize;
	}

	public Enums.ActionResult AddTask(Task task)
	{
		if(headOfTheList == null)
		{
			headOfTheList = task;
		}
		else
		{
			Task current = headOfTheList;
			while(current.GetNextTask()!= null && current.GetNextTask().GetDeadline()<task.GetDeadline())
			{
				current = current.GetNextTask();
			}
			if(current.GetDeadline()>task.GetDeadline())
			{
				headOfTheList = task;
				task.SetNextTask(current);
			}
			else
			{
				task.SetNextTask(current.GetNextTask());
				current.SetNextTask(task);
			}
		}

		taskSetSize++;
		totalUtilization += task.GetUtilisation(); 
		return Enums.ActionResult.Success;
	}

	public Enums.ActionResult AddTaskInInputOrder(Task task)
	{
		if(headOfTheList == null)
		{
			headOfTheList = task;
		}
		else
		{
			Task current = headOfTheList;

			while(current.GetNextTask()!= null)
			{
				current = current.GetNextTask();
			}

			current.SetNextTask(task);

		}

		taskSetSize++;
		totalUtilization += task.GetUtilisation(); 
		return Enums.ActionResult.Success;
	}

	public void Write()
	{
		System.out.println("Taskset elements " + taskSetSize+ ":");
		Task current = headOfTheList;
		while(current != null)
		{
			current.Print();
			current = current.GetNextTask();
		}
	}

	public int GetMaximumDeadline() 
	{
		Task task = headOfTheList;
		int maximumdeadline = task.GetDeadline(); 
		task = task.GetNextTask(); 

		while(task!=null)
		{
			if(maximumdeadline<task.GetDeadline())
				maximumdeadline = task.GetDeadline(); 
			task = task.GetNextTask(); 
		}

		return maximumdeadline;
	}

	public int GetMaximumPeriod() 
	{
		Task task = headOfTheList;
		int maximumperiod = task.GetMinInterArrival(); 
		task = task.GetNextTask(); 

		while(task!=null)
		{
			if(maximumperiod < task.GetMinInterArrival())
				maximumperiod = task.GetMinInterArrival();  
			task = task.GetNextTask(); 
		}

		return maximumperiod;
	}

	public int GetMinimumDeadline() 
	{
		Task task = headOfTheList;
		int minimumdeadline = task.GetDeadline(); 
		task = task.GetNextTask(); 

		while(task!=null)
		{
			if(minimumdeadline>task.GetDeadline())
				minimumdeadline = task.GetDeadline(); 
			task = task.GetNextTask(); 
		}

		return minimumdeadline;
	}

	public int GetMinimumPeriod() 
	{
		Task task = headOfTheList;
		int minimumperiod = task.GetMinInterArrival(); 
		task = task.GetNextTask(); 

		while(task!=null)
		{
			if(minimumperiod > task.GetMinInterArrival())
				minimumperiod = task.GetMinInterArrival();  
			task = task.GetNextTask(); 
		}

		return minimumperiod;
	}

	public void SetHeadOfTheList(Task t)
	{
		headOfTheList = t; 
	}

	public void SetTaskSetSize(int size)
	{
		taskSetSize = size;  
	}	
}